from abc import abstractmethod

from mex.common.connector import BaseConnector
from mex.common.identity.models import Identity
from mex.common.types import AnyMergedIdentifier, MergedPrimarySourceIdentifier


class BaseProvider(BaseConnector):
    """Base class to define the interface of identity providers."""

    @abstractmethod
    def assign(
        self,
        had_primary_source: MergedPrimarySourceIdentifier,
        identifier_in_primary_source: str,
    ) -> Identity:  # pragma: no cover
        """Find an Identity in a database or assign a new one."""
        ...

    @abstractmethod
    def fetch(
        self,
        *,
        had_primary_source: MergedPrimarySourceIdentifier | None = None,
        identifier_in_primary_source: str | None = None,
        stable_target_id: AnyMergedIdentifier | None = None,
    ) -> list[Identity]:  # pragma: no cover
        """Find Identity instances matching the given filters."""
        ...



import logging
import re
from enum import Enum
from typing import Any

import pytest
from click.testing import CliRunner
from pydantic import HttpUrl, create_model
from pydantic.fields import Field
from pytest import LogCaptureFixture

from mex.common.cli import _field_to_option, entrypoint
from mex.common.settings import BaseSettings


class MyStr(str):
    """Dummy string subclass for _field_to_option test."""

    __slots__ = ()


class MyEnum(Enum):
    """Dummy enum class for _field_to_option test."""

    FOO = 1
    BAR = 2


@pytest.mark.parametrize(
    ("name", "settings_cls", "info_dict"),
    [
        (
            "required_field",
            create_model(
                "RequiredFieldSettings",
                __base__=BaseSettings,
                required_field=(float, Field()),
            ),
            {
                "name": "required_field",
                "param_type_name": "option",
                "opts": ["--required-field"],
                "secondary_opts": [],
                "type": {"param_type": "Float", "name": "float"},
                "required": True,
                "nargs": 1,
                "multiple": False,
                "default": None,
                "envvar": "MEX_REQUIRED_FIELD",
                "help": None,
                "prompt": None,
                "is_flag": False,
                "flag_value": None,
                "count": False,
                "hidden": False,
            },
        ),
        (
            "url_field",
            create_model(
                "UrlFieldSettings",
                __base__=BaseSettings,
                url_field=(HttpUrl, "https://example.com"),
            ),
            {
                "name": "url_field",
                "param_type_name": "option",
                "opts": ["--url-field"],
                "secondary_opts": [],
                "type": {"param_type": "String", "name": "text"},
                "required": False,
                "nargs": 1,
                "multiple": False,
                "default": "https://example.com",
                "envvar": "MEX_URL_FIELD",
                "help": None,
                "prompt": None,
                "is_flag": False,
                "flag_value": None,
                "count": False,
                "hidden": False,
            },
        ),
        (
            "str_like_field",
            create_model(
                "StrLikeFieldSettings",
                __base__=BaseSettings,
                str_like_field=(MyStr, MyStr("test")),
            ),
            {
                "name": "str_like_field",
                "param_type_name": "option",
                "opts": ["--str-like-field"],
                "secondary_opts": [],
                "type": {"param_type": "String", "name": "text"},
                "required": False,
                "nargs": 1,
                "multiple": False,
                "default": "test",
                "envvar": "MEX_STR_LIKE_FIELD",
                "help": None,
                "prompt": None,
                "is_flag": False,
                "flag_value": None,
                "count": False,
                "hidden": False,
            },
        ),
        (
            "optional_flag",
            create_model(
                "OptionalFlagSettings",
                __base__=BaseSettings,
                optional_flag=(
                    bool,
                    Field(default=False, description="This flag is optional"),
                ),
            ),
            {
                "name": "optional_flag",
                "param_type_name": "option",
                "opts": ["--optional-flag"],
                "secondary_opts": [],
                "type": {"param_type": "Bool", "name": "boolean"},
                "required": False,
                "nargs": 1,
                "multiple": False,
                "default": False,
                "envvar": "MEX_OPTIONAL_FLAG",
                "help": "This flag is optional",
                "prompt": None,
                "is_flag": True,
                "flag_value": True,
                "count": False,
                "hidden": False,
            },
        ),
        (
            "enum_list",
            create_model(
                "EnumListSettings",
                __base__=BaseSettings,
                enum_list=(
                    list[MyEnum],
                    Field(
                        [MyEnum.FOO, MyEnum.BAR], description="Multiple values allowed"
                    ),
                ),
            ),
            {
                "name": "enum_list",
                "param_type_name": "option",
                "opts": ["--enum-list"],
                "secondary_opts": [],
                "type": {"name": "text", "param_type": "String"},
                "required": False,
                "nargs": 1,
                "multiple": False,  # on purpose, because we let pydantic parse lists
                "default": "[1, 2]",
                "envvar": "MEX_ENUM_LIST",
                "help": "Multiple values allowed",
                "prompt": None,
                "is_flag": False,
                "flag_value": None,
                "count": False,
                "hidden": False,
            },
        ),
        (
            "union_field",
            create_model(
                "UnionFieldSettings",
                __base__=BaseSettings,
                union_field=(
                    bool | str,
                    Field(default=True, description="String or boolean"),
                ),
            ),
            {
                "name": "union_field",
                "param_type_name": "option",
                "opts": ["--union-field"],
                "secondary_opts": [],
                "type": {"name": "text", "param_type": "String"},
                "required": False,
                "nargs": 1,
                "multiple": False,
                "default": "true",
                "envvar": "MEX_UNION_FIELD",
                "help": "String or boolean",
                "prompt": None,
                "is_flag": False,
                "flag_value": None,
                "count": False,
                "hidden": False,
            },
        ),
    ],
    ids=[
        "required field",
        "url field",
        "str-like field",
        "optional flag",
        "enum list",
        "union field",
    ],
)
def test_field_to_option(
    name: str, settings_cls: type[BaseSettings], info_dict: dict[str, Any]
) -> None:
    option = _field_to_option(name, settings_cls)
    assert option.to_info_dict() == info_dict


def test_good_entrypoint_exits_zero() -> None:
    @entrypoint(BaseSettings)
    def good_entrypoint() -> None:
        return

    result = CliRunner().invoke(good_entrypoint, args=[])
    assert result.exit_code == 0, result.stdout


def test_faulty_entrypoint_exits_non_zero() -> None:
    @entrypoint(BaseSettings)
    def faulty_entrypoint() -> None:
        _ = 1 / 0

    result = CliRunner().invoke(faulty_entrypoint, args=[])
    assert result.exit_code == 1, result.stdout


class ChattySettings(BaseSettings):
    custom_setting: str = "default"


def test_entrypoint_logs_docs_and_settings(caplog: LogCaptureFixture) -> None:
    @entrypoint(ChattySettings)
    def chatty_entrypoint() -> None:
        """Hi, I am Pointy McEntryFace."""
        return

    with caplog.at_level(logging.INFO, logger="mex"):
        result = CliRunner().invoke(
            chatty_entrypoint, args=["--custom-setting=override"]
        )
    assert result.exit_code == 0, result.stdout

    assert "Pointy McEntryFace" in caplog.text
    assert re.search(r"custom_setting\s+override", caplog.text)



import pytest
from requests import HTTPError

from mex.common.orcid.connector import OrcidConnector
from mex.common.orcid.models import (
    OrcidIdentifier,
    OrcidSearchItem,
    OrcidSearchResponse,
)


@pytest.mark.usefixtures("mocked_orcid")
@pytest.mark.parametrize(
    ("filters", "expected"),
    [
        (
            {"given-names": "Josiah", "family-name": "Carberry"},
            "given-names:Josiah AND family-name:Carberry",
        ),
        (
            {"given-and-family-names": '"Jayne Carberry"'},
            'given-and-family-names:"Jayne Carberry"',
        ),
        (
            {"givennames": "Josiah", "familyname": "Carberry"},
            "givennames:Josiah AND familyname:Carberry",
        ),
    ],
    ids=["valid_query", "valid given-and_family-names query", "non_valid_query"],
)
def test_build_query(filters: dict[str, str], expected: str) -> None:
    orcid_api = OrcidConnector.get()
    built_query = orcid_api.build_query(filters=filters)
    assert built_query == expected


@pytest.mark.usefixtures("orcid_person_raw")
@pytest.mark.parametrize(
    ("family_name", "given_names", "expected"),
    [
        (
            "Doe",
            "John",
            OrcidSearchResponse(
                num_found=1,
                result=[
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0009-0004-3041-5706",
                            uri="https://orcid.org/0009-0004-3041-5706",
                        )
                    )
                ],
            ),
        ),
        (
            "Doe",
            "Multiple",
            OrcidSearchResponse(
                num_found=10,
                result=[
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0009-0005-5828-7053",
                            uri="https://orcid.org/0009-0005-5828-7053",
                        )
                    ),
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0000-0003-3648-8952",
                            uri="https://orcid.org/0000-0003-3648-8952",
                        )
                    ),
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0000-0002-7523-2549",
                            uri="https://orcid.org/0000-0002-7523-2549",
                        )
                    ),
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0000-0002-9056-5667",
                            uri="https://orcid.org/0000-0002-9056-5667",
                        )
                    ),
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0000-0002-3372-2005",
                            uri="https://orcid.org/0000-0002-3372-2005",
                        )
                    ),
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0000-0001-7659-8932",
                            uri="https://orcid.org/0000-0001-7659-8932",
                        )
                    ),
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0009-0005-0959-5447",
                            uri="https://orcid.org/0009-0005-0959-5447",
                        )
                    ),
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0009-0000-4002-171X",
                            uri="https://orcid.org/0009-0000-4002-171X",
                        )
                    ),
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0009-0006-0442-1402",
                            uri="https://orcid.org/0009-0006-0442-1402",
                        )
                    ),
                    OrcidSearchItem(
                        orcid_identifier=OrcidIdentifier(
                            path="0009-0006-9954-421X",
                            uri="https://orcid.org/0009-0006-9954-421X",
                        )
                    ),
                ],
            ),
        ),
        ("Doe", "NotExistJohn", OrcidSearchResponse(num_found=0, result=[])),
    ],
    ids=["existing person", "multiple results", "non-existing person"],
)
@pytest.mark.usefixtures("mocked_orcid")
def test_fetch_person_by_name(
    family_name: str, given_names: str, expected: OrcidSearchResponse
) -> None:
    connector = OrcidConnector.get()
    response = connector.search_records_by_name(given_names, family_name)
    assert response == expected


@pytest.mark.usefixtures("mocked_orcid")
def test_get_record_by_id() -> None:
    connector = OrcidConnector.get()
    result = connector.get_record_by_id("0009-0004-3041-5706")
    assert result.model_dump() == {
        "orcid_identifier": {
            "path": "0009-0004-3041-5706",
            "uri": "https://orcid.org/0009-0004-3041-5706",
        },
        "person": {
            "emails": {"email": []},
            "name": {
                "family_name": {"value": "Doe"},
                "given_names": {"value": "John"},
                "visibility": "public",
            },
        },
    }


@pytest.mark.usefixtures("mocked_orcid")
def test_get_record_by_id_not_found() -> None:
    connector = OrcidConnector.get()
    with pytest.raises(HTTPError, match="404 Not Found"):
        connector.get_record_by_id("0000-0000-0000-000")



from typing import Annotated, ClassVar, Literal

from pydantic import Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    ExtractedVariableGroupIdentifier,
    MergedPrimarySourceIdentifier,
    MergedResourceIdentifier,
    MergedVariableGroupIdentifier,
    Text,
)


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["VariableGroup"], Field(frozen=True)]] = (
        "VariableGroup"
    )


class _RequiredLists(_Stem):
    containedBy: Annotated[
        list[MergedResourceIdentifier],
        Field(
            min_length=1,
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/isPartOf"]},
        ),
    ]
    label: Annotated[
        list[Text],
        Field(
            min_length=1,
            json_schema_extra={
                "sameAs": ["http://www.w3.org/2000/01/rdf-schema#label"]
            },
        ),
    ]


class _SparseLists(_Stem):
    containedBy: Annotated[
        list[MergedResourceIdentifier],
        Field(
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/isPartOf"]}
        ),
    ] = []
    label: Annotated[
        list[Text],
        Field(
            json_schema_extra={"sameAs": ["http://www.w3.org/2000/01/rdf-schema#label"]}
        ),
    ] = []


class BaseVariableGroup(
    _RequiredLists,
    json_schema_extra={
        "description": (
            "The grouping of variables according to a certain aspect, e.g. how the "
            "information is modelled in the primary source."
        ),
        "title": "Variable Group",
    },
):
    """All fields for a valid variable group except for provenance."""


class ExtractedVariableGroup(BaseVariableGroup, ExtractedData):
    """An automatically extracted metadata set describing a variable group."""

    entityType: Annotated[
        Literal["ExtractedVariableGroup"], Field(alias="$type", frozen=True)
    ] = "ExtractedVariableGroup"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(
        self,
    ) -> Annotated[
        ExtractedVariableGroupIdentifier,
        Field(
            json_schema_extra={"sameAs": ["http://purl.org/dc/elements/1.1/identifier"]}
        ),
    ]:
        """Return the computed identifier for this extracted item."""
        return self._get_identifier(ExtractedVariableGroupIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(self) -> MergedVariableGroupIdentifier:  # noqa: N802
        """Return the computed stableTargetId for this extracted item."""
        return self._get_stable_target_id(MergedVariableGroupIdentifier)


class MergedVariableGroup(BaseVariableGroup, MergedItem):
    """The result of merging all extracted items and rules for a variable group."""

    entityType: Annotated[
        Literal["MergedVariableGroup"], Field(alias="$type", frozen=True)
    ] = "MergedVariableGroup"
    identifier: Annotated[MergedVariableGroupIdentifier, Field(frozen=True)]


class PreviewVariableGroup(_SparseLists, PreviewItem):
    """Preview for merging all extracted items and rules for a variable group."""

    entityType: Annotated[
        Literal["PreviewVariableGroup"], Field(alias="$type", frozen=True)
    ] = "PreviewVariableGroup"
    identifier: Annotated[MergedVariableGroupIdentifier, Field(frozen=True)]


class AdditiveVariableGroup(_SparseLists, AdditiveRule):
    """Rule to add values to merged variable group items."""

    entityType: Annotated[
        Literal["AdditiveVariableGroup"], Field(alias="$type", frozen=True)
    ] = "AdditiveVariableGroup"


class SubtractiveVariableGroup(_SparseLists, SubtractiveRule):
    """Rule to subtract values from merged variable group items."""

    entityType: Annotated[
        Literal["SubtractiveVariableGroup"], Field(alias="$type", frozen=True)
    ] = "SubtractiveVariableGroup"


class PreventiveVariableGroup(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged variable group items."""

    entityType: Annotated[
        Literal["PreventiveVariableGroup"], Field(alias="$type", frozen=True)
    ] = "PreventiveVariableGroup"
    containedBy: list[MergedPrimarySourceIdentifier] = []
    label: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    additive: AdditiveVariableGroup = AdditiveVariableGroup()
    subtractive: SubtractiveVariableGroup = SubtractiveVariableGroup()
    preventive: PreventiveVariableGroup = PreventiveVariableGroup()


class VariableGroupRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update a variable group item."""

    entityType: Annotated[
        Literal["VariableGroupRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "VariableGroupRuleSetRequest"


class VariableGroupRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve a variable group item."""

    entityType: Annotated[
        Literal["VariableGroupRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "VariableGroupRuleSetResponse"
    stableTargetId: MergedVariableGroupIdentifier


class VariableGroupMapping(_Stem, BaseMapping):
    """Mapping for describing a variable group transformation."""

    entityType: Annotated[
        Literal["VariableGroupMapping"], Field(alias="$type", frozen=True)
    ] = "VariableGroupMapping"
    hadPrimarySource: Annotated[
        list[MappingField[MergedPrimarySourceIdentifier]], Field(min_length=1)
    ]
    identifierInPrimarySource: Annotated[list[MappingField[str]], Field(min_length=1)]
    containedBy: Annotated[
        list[MappingField[list[MergedResourceIdentifier]]], Field(min_length=1)
    ]
    label: Annotated[list[MappingField[list[Text]]], Field(min_length=1)]


class VariableGroupFilter(_Stem, BaseFilter):
    """Class for defining filter rules for variable group items."""

    entityType: Annotated[
        Literal["VariableGroupFilter"], Field(alias="$type", frozen=True)
    ] = "VariableGroupFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



import json
import re
from collections.abc import Iterable
from enum import Enum
from functools import lru_cache
from pathlib import PurePath
from typing import Any
from uuid import UUID

from pydantic import AnyUrl, SecretStr
from pydantic import BaseModel as PydanticModel


class MExEncoder(json.JSONEncoder):
    """Custom JSON encoder that can handle pydantic models, enums and UUIDs."""

    def default(self, obj: object) -> object:  # noqa: PLR0911
        """Implement custom serialization rules."""
        # break import cycle, sigh
        from mex.common.types import PathWrapper, TemporalEntity  # noqa: PLC0415

        if isinstance(obj, PydanticModel):
            return obj.model_dump()
        if isinstance(obj, AnyUrl):
            return str(obj)
        if isinstance(obj, SecretStr):
            return obj.get_secret_value()
        if isinstance(obj, Enum):
            return obj.value
        if isinstance(obj, UUID):
            return str(obj)
        if isinstance(obj, TemporalEntity):
            return str(obj)
        if isinstance(obj, PurePath):
            return obj.as_posix()
        if isinstance(obj, PathWrapper):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


@lru_cache(maxsize=1024)
def snake_to_dromedary(string: str) -> str:
    """Convert the given string from `snake_case` into `dromedaryCase`."""
    if len(tokens := re.split(r"_", string)) > 1:
        return "".join(
            word.capitalize() if index else word.lower()
            for index, word in enumerate(tokens)
        )
    return string


@lru_cache(maxsize=1024)
def dromedary_to_snake(string: str) -> str:
    """Convert the given string from `dromedaryCase` into `snake_case`."""
    return "_".join(
        word.lower()
        for word in re.split(r"([A-Z]+(?![a-z])|[a-z]+|[A-Z][a-z]+)", string)
        if word.strip("_")
    )


@lru_cache(maxsize=1024)
def dromedary_to_kebab(string: str) -> str:
    """Convert the given string from `dromedaryCase` into `kebab-case`."""
    return "-".join(
        word.lower()
        for word in re.split(r"([A-Z]+(?![a-z])|[a-z]+|[A-Z][a-z]+)", string)
        if word.strip("-")
    )


@lru_cache(maxsize=1024)
def kebab_to_camel(string: str) -> str:
    """Convert the given string from `kebab-case` into `CamelCase`."""
    if len(tokens := re.split(r"\-+", string)) > 1:
        return "".join(word.title() for word in tokens)
    return string[:1].upper() + string[1:]


@lru_cache(maxsize=1024)
def normalize(string: str) -> str:
    """Normalize the given string to lowercase, numerals and single spaces."""
    return " ".join(re.sub(r"[^a-z0-9]", " ", string.lower()).split())


@lru_cache(maxsize=1024)
def split_to_caps(string: str) -> str:
    """Convert the given string from `Split case` into `CAPS_CASE`."""
    return "_".join(word.upper() for word in re.split("[^a-zA-Z]", string) if word)


def ensure_prefix(string_like: object, prefix: object) -> str:
    """Return a string with the given prefix prepended if it is not present yet.

    If `string_like` already starts with the prefix, return a stringified copy.
    This method is the inverse of `str.removeprefix`.

    Args:
        string_like: Object to convert to string and potentially prefix.
        prefix: Object to convert to string and use as prefix.

    Returns:
        String with the prefix guaranteed to be present at the beginning.
    """
    string = str(string_like)
    prefix = str(prefix)
    if string.startswith(prefix):
        return string
    return f"{prefix}{string}"


def ensure_postfix(string_like: object, postfix: object) -> str:
    """Return a string with the given postfix appended if it is not present yet.

    If `string_like` already ends with the postfix, return a stringified copy.
    This method is the inverse of `str.removepostfix`.

    Args:
        string_like: Object to convert to string and potentially postfix.
        postfix: Object to convert to string and use as postfix.

    Returns:
        String with the postfix guaranteed to be present at the end.
    """
    string = str(string_like)
    postfix = str(postfix)
    if string.endswith(postfix):
        return string
    return f"{string}{postfix}"


def to_key_and_values(dct: dict[str, Any]) -> Iterable[tuple[str, list[Any]]]:
    """Return an iterable of dictionary items where the values are always lists.

    Normalizes dictionary values by converting single values to single-item lists,
    leaving existing lists unchanged, and converting None to empty lists.

    Args:
        dct: Dictionary to normalize the values of.

    Yields:
        Tuples of (key, list_value) where list_value is guaranteed to be a list.
    """
    for key, value in dct.items():
        if value is None:
            list_of_values = []
        elif isinstance(value, list):
            list_of_values = value
        else:
            list_of_values = [value]
        yield key, list_of_values



import logging
from collections.abc import Generator
from uuid import UUID

from pytest import LogCaptureFixture

from mex.common.exceptions import MExError
from mex.common.logging import watch


def test_watch(caplog: LogCaptureFixture) -> None:
    items: list[str | UUID | MExError] = [
        "foo",
        UUID(int=16, version=4),
        MExError("foo", 42),
    ]

    @watch(log_interval=1)
    def dummy_generator() -> Generator[str | UUID | MExError, None, None]:
        yield from items

    # unpack generator and capture logs
    with caplog.at_level(logging.INFO, logger="mex"):
        yielded_items = list(dummy_generator())

    # check that returned items are untouched
    assert yielded_items == items

    # check captured logging
    assert len(caplog.messages) == 3
    str_line, uuid_line, error_line = caplog.messages
    assert "dummy_generator - foo" in str_line
    assert "dummy_generator - 00000000-0000-4000-8000-000000000010" in uuid_line
    assert "dummy_generator - MExError: foo, 42" in error_line



from enum import Enum, EnumMeta
from typing import TYPE_CHECKING, ClassVar, Self, Union

from pydantic import (
    AnyUrl,
    BaseModel,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    json_schema,
)
from pydantic_core import core_schema

from mex.common.transform import normalize, split_to_caps
from mex.model import VOCABULARY_JSON_BY_NAME

if TYPE_CHECKING:  # pragma: no cover
    from enum import _EnumDict

    from mex.common.types import Text

VOCABULARY_PATTERN = r"https://mex.rki.de/item/[a-z0-9-]+"


class BilingualText(BaseModel):
    """String-field translated in German and English."""

    de: str | None = None
    en: str | None = None


class Concept(BaseModel):
    """Single entry in a vocabulary with stable ID, labels and definition."""

    identifier: AnyUrl
    inScheme: AnyUrl
    prefLabel: BilingualText
    altLabel: list[BilingualText] = []
    definition: BilingualText | None = None


class VocabularyLoader(EnumMeta):
    """Metaclass to load names and values from a JSON file and create a dynamic enum."""

    def __new__(
        cls, name: str, bases: tuple[type], dct: "_EnumDict"
    ) -> "VocabularyLoader":
        """Create a new enum class by loading the configured vocabulary JSON."""
        if vocabulary_name := dct.get("__vocabulary__"):
            dct["__concepts__"] = cls.parse_raw(vocabulary_name.replace("-", "_"))
            for concept in dct["__concepts__"]:
                dct[split_to_caps(concept.prefLabel.en)] = str(concept.identifier)
        return super().__new__(cls, name, bases, dct)

    @classmethod
    def parse_raw(cls, vocabulary_name: str) -> list[Concept]:
        """Parse vocabulary and return concepts as list."""
        return [
            Concept.model_validate(raw_vocabulary)
            for raw_vocabulary in VOCABULARY_JSON_BY_NAME[vocabulary_name]
        ]


class VocabularyEnum(Enum, metaclass=VocabularyLoader):
    """Base class for vocabulary enums that sets the correct metaclass."""

    __vocabulary__: ClassVar[str]
    __concepts__: ClassVar[list[Concept]]

    @classmethod
    def find(cls, search_term: Union[str, "Text"]) -> Self | None:
        """Get the enum instance that matches a label of the underlying concepts.

        The given `search_term` can be string or a Text with an optional language
        setting to narrow down the search fields.
        The `prefLabel` and `altLabel` of the concepts which were used to create this
        vocabulary are searched for exact matches to the `search_term`.

        Args:
            search_term: String or Text to look for

        Returns:
            Enum instance for the found concept or None
        """
        language = getattr(search_term, "language", None)
        search_term = normalize(str(search_term))
        for concept in cls.__concepts__:
            searchable_labels = []
            for label in (concept.prefLabel, *concept.altLabel):
                if not label:
                    continue
                if language is None:
                    if label.de:
                        searchable_labels.append(normalize(label.de))
                    if label.en:
                        searchable_labels.append(normalize(label.en))
                elif language_label := label.model_dump().get(language.value):
                    searchable_labels.append(normalize(language_label))
            if search_term in searchable_labels:
                return cls(str(concept.identifier))
        return None

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: object, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """Modify the core schema to add the vocabulary regex."""
        return core_schema.json_or_python_schema(
            json_schema=core_schema.union_schema(
                [
                    core_schema.str_schema(pattern=VOCABULARY_PATTERN),
                    core_schema.no_info_plain_validator_function(cls),
                ],
            ),
            python_schema=core_schema.chain_schema(
                [
                    core_schema.is_instance_schema(cls | str),
                    core_schema.no_info_plain_validator_function(cls),
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda s: s.value,
                when_used="unless-none",
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema_: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> json_schema.JsonSchemaValue:
        """Modify the json schema to add the scheme and an example."""
        json_schema_ = handler(core_schema_)
        json_schema_["examples"] = [str(cls.__concepts__[0].identifier)]
        json_schema_["useScheme"] = f"https://mex.rki.de/item/{cls.__vocabulary__}"
        return json_schema_

    def __repr__(self) -> str:
        """Overwrite representation because dynamic enum names are unknown to mypy."""
        return f'{self.__class__.__name__}["{self.name}"]'


class AccessRestriction(VocabularyEnum):
    """The access restriction type."""

    __vocabulary__ = "access-restriction"


class ActivityType(VocabularyEnum):
    """The activity type."""

    __vocabulary__ = "activity-type"


class AnonymizationPseudonymization(VocabularyEnum):
    """Whether the resource is anonymized/pseudonymized."""

    __vocabulary__ = "anonymization-pseudonymization"


class APIType(VocabularyEnum):
    """Technical standard or style of a network API."""

    __vocabulary__ = "api-type"


class BibliographicResourceType(VocabularyEnum):
    """The type of a bibliographic resource."""

    __vocabulary__ = "bibliographic-resource-type"


class ConsentStatus(VocabularyEnum):
    """The status of a consent."""

    __vocabulary__ = "consent-status"


class ConsentType(VocabularyEnum):
    """The type of a consent."""

    __vocabulary__ = "consent-type"


class DataProcessingState(VocabularyEnum):
    """Type for state of data processing."""

    __vocabulary__ = "data-processing-state"


class Frequency(VocabularyEnum):
    """Frequency type."""

    __vocabulary__ = "frequency"


class Language(VocabularyEnum):
    """Language type."""

    __vocabulary__ = "language"


class License(VocabularyEnum):
    """License type."""

    __vocabulary__ = "license"


class MIMEType(VocabularyEnum):
    """The mime type."""

    __vocabulary__ = "mime-type"


class PersonalData(VocabularyEnum):
    """Classification of personal data."""

    __vocabulary__ = "personal-data"


class ResourceCreationMethod(VocabularyEnum):
    """The creation method of a resource."""

    __vocabulary__ = "resource-creation-method"


class ResourceTypeGeneral(VocabularyEnum):
    """The general type of a resource."""

    __vocabulary__ = "resource-type-general"


class TechnicalAccessibility(VocabularyEnum):
    """Technical accessibility within RKI and outside of RKI."""

    __vocabulary__ = "technical-accessibility"


class Theme(VocabularyEnum):
    """The theme type."""

    __vocabulary__ = "theme"



from typing import Generic, TypeVar

from pydantic import BaseModel

_ContainerItemT = TypeVar("_ContainerItemT")


class ItemsContainer(BaseModel, Generic[_ContainerItemT]):
    """Generic container that contains items."""

    items: list[_ContainerItemT]


class PaginatedItemsContainer(BaseModel, Generic[_ContainerItemT]):
    """Generic container that contains items and has a total item count."""

    items: list[_ContainerItemT]
    total: int



import pytest
from pydantic import BaseModel, ValidationError

from mex.common.types import Text
from mex.common.types.text import (
    TextLanguage,
    get_language_by_confidence,
)


def test_text_language_detect() -> None:
    # accepted languages are only German and English. Other languages are None.
    de_text = Text(value="Diese tiefen Seufzern haben einen Sinn. Legt sie uns aus.")
    assert de_text.language == TextLanguage.DE
    en_text = Text(value="There's matter in these sighs. You must translate.")
    assert en_text.language == TextLanguage.EN

    # language that can be detected by langdetect but should be excluded from results
    fr_text = Text(value="Ces profonds soupirs ont un sens. Expliquez-les-nous.")
    assert fr_text.language is None
    es_text = Text(
        value="Estos profundos suspiros tienen un significado. Tradúcelos para nosotros."
    )
    assert es_text.language is None
    ru_text = Text(value="Эти глубокие вздохи имеют смысл. Объясните их нам.")
    assert ru_text.language is None
    ko_text = Text(value="이 한숨에는 문제가 있습니다. 번역해야 합니다.")
    assert ko_text.language is None

    # language that langdetect does not know and should not be included in results
    am_text = Text(value="በእነዚህ ቅስቀሳዎች ውስጥ ቁም ነገር አለ። መተርጎም አለብዎት.")
    assert am_text.language is None

    # check that we can overwrite the language
    overwrite_text = Text(
        value="There's matter in these sighs. You must translate",
        language=TextLanguage.DE,
    )
    assert overwrite_text.language == TextLanguage.DE

    # check that we can explicitly null the language
    none_text = Text(
        value="There's matter in these sighs. You must translate", language=None
    )
    assert none_text.language is None


class DummyModel(BaseModel):
    text: Text


def test_text_validation() -> None:
    with pytest.raises(ValidationError, match="Allowed input types are dict and str"):
        _ = DummyModel.model_validate({"text": 1})

    model = DummyModel.model_validate({"text": "we are parsing a string here"})
    assert model.model_dump() == {
        "text": {
            "value": "we are parsing a string here",
            "language": TextLanguage.EN,
        }
    }

    model = DummyModel.model_validate(
        {"text": {"value": "and here, we are parsing an object"}}
    )
    assert model.model_dump() == {
        "text": {
            "value": "and here, we are parsing an object",
            "language": TextLanguage.EN,
        }
    }

    model = DummyModel.model_validate(
        {
            "text": {
                "value": "now we parse an object with a language key",
                "language": "de",
            }
        }
    )
    assert model.model_dump() == {
        "text": {
            "value": "now we parse an object with a language key",
            "language": TextLanguage.DE,
        }
    }


def test_text_hash() -> None:
    text = Text(value="Hallo Welt.", language=TextLanguage.DE)
    assert hash(text) == hash(("Hallo Welt.", TextLanguage.DE))


class Prob:
    def __init__(self, lang: str, prob: float) -> None:
        self.lang = lang
        self.prob = prob


class MockedDetector:
    def __init__(self, probs: list[Prob]) -> None:
        self._probs = probs

    def get_probabilities(self) -> list[Prob]:
        return self._probs


@pytest.mark.parametrize(
    ("probs", "expected"),
    [
        ([Prob("en", 0.95)], TextLanguage.EN),
        ([Prob("de", 0.88)], TextLanguage.DE),
        ([Prob("fr", 0.92)], None),
        ([Prob("en", 0.40)], None),
        (
            [Prob("de", 0.40), Prob("en", 0.20), Prob("de", 0.80)],
            TextLanguage.DE,
        ),
        ([], None),
    ],
    ids=[
        "high_confidence_en",
        "high_confidence_de",
        "fr_unsupported_none",
        "low_confidence_en_none",
        "highest_prob_selected_de",
        "no_language_detected_none",
    ],
)
def test_get_language_by_confidence(
    probs: list[Prob],
    expected: TextLanguage | None,
) -> None:
    detector = MockedDetector(probs)
    result = get_language_by_confidence(detector)
    assert result == expected



import hashlib

from mex.common.identity.base import BaseProvider
from mex.common.identity.models import Identity
from mex.common.models import (
    MEX_PRIMARY_SOURCE_IDENTIFIER,
    MEX_PRIMARY_SOURCE_IDENTIFIER_IN_PRIMARY_SOURCE,
    MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
)
from mex.common.types import (
    AnyMergedIdentifier,
    Identifier,
    MergedPrimarySourceIdentifier,
)


class MemoryIdentityProvider(BaseProvider):
    """Connector class to handle read/write to the identity database."""

    def __init__(self) -> None:
        """Initialize an in-memory database with the identity of MEx itself."""
        self._database: list[Identity] = [
            Identity(
                identifier=MEX_PRIMARY_SOURCE_IDENTIFIER,
                hadPrimarySource=MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
                identifierInPrimarySource=MEX_PRIMARY_SOURCE_IDENTIFIER_IN_PRIMARY_SOURCE,
                stableTargetId=MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
            )
        ]

    @staticmethod
    def _get_identifier(*args: str) -> Identifier:
        """Get deterministic identifier based on args.

        Creates a deterministic identifier by joining the arguments with newlines,
        computing an MD5 hash, and using the resulting integer as a seed for
        identifier generation.

        Args:
            *args: String arguments to use for identifier generation.

        Returns:
            Deterministic Identifier based on the input arguments.
        """
        seed_string = "\n".join(args)
        hash_ = hashlib.md5(  # noqa: S324 identifier generation is not security related
            seed_string.encode()
        )
        seed_hex = hash_.hexdigest()
        seed_int = int(seed_hex, 16)
        return Identifier.generate(seed=seed_int)

    def assign(
        self,
        had_primary_source: MergedPrimarySourceIdentifier,
        identifier_in_primary_source: str,
    ) -> Identity:
        """Find an Identity in the in-memory database or assign a new one.

        Args:
            had_primary_source: Stable target ID of primary source
            identifier_in_primary_source: Identifier in the primary source

        Returns:
            Newly created or updated Identity instance
        """
        identities = self.fetch(
            had_primary_source=had_primary_source,
            identifier_in_primary_source=identifier_in_primary_source,
        )
        if identities:
            return identities[0]

        identity = Identity(
            hadPrimarySource=had_primary_source,
            identifierInPrimarySource=identifier_in_primary_source,
            stableTargetId=self._get_identifier(
                "stableTargetId", had_primary_source, identifier_in_primary_source
            ),
            identifier=self._get_identifier(
                "identifier", had_primary_source, identifier_in_primary_source
            ),
        )
        self._database.append(identity)
        return identity

    def fetch(
        self,
        *,
        had_primary_source: MergedPrimarySourceIdentifier | None = None,
        identifier_in_primary_source: str | None = None,
        stable_target_id: AnyMergedIdentifier | None = None,
    ) -> list[Identity]:
        """Find Identity instances in the in-memory database.

        Args:
            had_primary_source: Stable target ID of primary source
            identifier_in_primary_source: Identifier in the primary source
            stable_target_id: Stable target ID of the entity

        Returns:
            List of Identity instances
        """
        identities = iter(self._database)

        if had_primary_source:
            identities = filter(
                lambda i: i.hadPrimarySource == had_primary_source, identities
            )
        if identifier_in_primary_source:
            identities = filter(
                lambda i: i.identifierInPrimarySource == identifier_in_primary_source,
                identities,
            )
        if stable_target_id:
            identities = filter(
                lambda i: i.stableTargetId == stable_target_id, identities
            )

        return list(identities)

    def metrics(self) -> dict[str, int]:
        """Generate metrics about identity provider usage."""
        return {"database_size": len(self._database)}

    def close(self) -> None:
        """Trash the in-memory identity database."""
        self._database.clear()



from pathlib import Path

import pytest

from mex.common.types import PathWrapper


def test_path_wrapper_instantiation() -> None:
    assert PathWrapper(Path("foo", "bar"))._path == Path("foo", "bar")
    assert PathWrapper("foo/bar")._path == Path("foo", "bar")
    assert PathWrapper(PathWrapper("foo/bar"))._path == Path("foo", "bar")


def test_path_wrapper_fspath() -> None:
    assert (
        PathWrapper(Path("foo") / "bar").__fspath__() == Path("foo", "bar").__fspath__()
    )


def test_path_wrapper_slash() -> None:
    assert PathWrapper(Path("foo")) / "bar" == PathWrapper(Path("foo") / "bar")._path


def test_path_wrapper_string() -> None:
    assert str(PathWrapper(Path("foo", "bar"))) == Path("foo", "bar").as_posix()


def test_path_wrapper_representation() -> None:
    assert repr(PathWrapper(Path("foo", "bar"))).startswith("PathWrapper")


def test_path_wrapper_equality() -> None:
    assert PathWrapper(Path("foo", "bar")) == PathWrapper(Path("foo", "bar"))
    assert PathWrapper(Path("foo", "bar")) != PathWrapper(Path("bar", "batz"))

    with pytest.raises(TypeError):
        _ = PathWrapper("foo") == 42


def test_path_wrapper_relative_absolute() -> None:
    assert PathWrapper("/foo/bar").is_absolute() == Path("/foo/bar").is_absolute()
    assert (
        PathWrapper("C:\\foo\\bar").is_absolute() == Path("C:\\foo\\bar").is_absolute()
    )
    assert PathWrapper("bar/batz").is_relative() != Path("bar/batz").is_absolute()



from mex.common.identity.base import BaseProvider
from mex.common.identity.models import Identity
from mex.common.identity.registry import get_provider, register_provider

__all__ = (
    "BaseProvider",
    "Identity",
    "get_provider",
    "register_provider",
)



from pathlib import Path
from unittest.mock import MagicMock

import pytest
import requests
from pytest import MonkeyPatch

from mex.common.settings import BaseSettings
from mex.common.wikidata.connector import WikidataAPIConnector

TESTDATA_DIR = Path(__file__).parent / "test_data"


@pytest.fixture
def mocked_session_wikidata_api(monkeypatch: MonkeyPatch) -> MagicMock:
    """Mock and return WikidataAPIConnector with a MagicMock session."""
    mocked_session = MagicMock(spec=requests.Session)

    def __init__(self: WikidataAPIConnector) -> None:
        self.session = mocked_session

    monkeypatch.setattr(WikidataAPIConnector, "__init__", __init__)
    return mocked_session


@pytest.fixture(autouse=True)
def set_wiki_api_url_for_integration_tests(
    monkeypatch: MonkeyPatch,
    is_integration_test: bool,  # noqa: FBT001
) -> None:
    """Set correct Wikidata API URL for integration tests."""
    if is_integration_test:
        settings = BaseSettings.get()
        monkeypatch.setattr(
            settings,
            "wiki_api_url",
            "https://www.wikidata.org/w/api.php",
        )



import pytest

from mex.common.orcid.models import (
    OrcidEmail,
    OrcidEmails,
    OrcidFamilyName,
    OrcidGivenNames,
    OrcidIdentifier,
    OrcidName,
    OrcidPerson,
    OrcidRecord,
)


@pytest.mark.parametrize(
    ("orcid_data", "expected_result"),
    [
        (
            {
                "orcid_identifier": {
                    "path": "0009-0004-3041-5706",
                    "uri": "https://orcid.org/0009-0004-3041-5706",
                },
                "person": {
                    "emails": {"email": [{"email": "test@example.com"}]},
                    "name": {
                        "family_name": {"value": "Carberry"},
                        "given_names": {"value": "Josiah"},
                        "visibility": "public",
                    },
                },
            },
            OrcidRecord(
                orcid_identifier=OrcidIdentifier(
                    path="0009-0004-3041-5706",
                    uri="https://orcid.org/0009-0004-3041-5706",
                ),
                person=OrcidPerson(
                    emails=OrcidEmails(email=[OrcidEmail(email="test@example.com")]),
                    name=OrcidName(
                        family_name=OrcidFamilyName(value="Carberry"),
                        given_names=OrcidGivenNames(value="Josiah"),
                        visibility="public",
                    ),
                ),
            ),
        ),
        (
            {
                "orcid-identifier": {
                    "path": "0009-0004-3041-5706",
                    "uri": "https://orcid.org/0009-0004-3041-5706",
                },
                "person": {
                    "name": {
                        "given-names": {"value": "Josiah"},
                        "family-name": {"value": "Carberry"},
                        "visibility": "public",
                    },
                    "emails": {"email": []},
                },
            },
            OrcidRecord(
                orcid_identifier=OrcidIdentifier(
                    path="0009-0004-3041-5706",
                    uri="https://orcid.org/0009-0004-3041-5706",
                ),
                person=OrcidPerson(
                    emails=OrcidEmails(email=[]),
                    name=OrcidName(
                        family_name=OrcidFamilyName(value="Carberry"),
                        given_names=OrcidGivenNames(value="Josiah"),
                        visibility="public",
                    ),
                ),
            ),
        ),
        (
            {
                "orcid-identifier": {
                    "path": "0009-0004-3041-5706",
                    "uri": "https://orcid.org/0009-0004-3041-5706",
                },
                "person": {
                    "name": {
                        "given-names": {"value": "Josiah"},
                        "family-name": {"value": "Carberry"},
                        "visibility": "private",
                    },
                    "emails": {"email": [{"email": "test@example.com"}]},
                },
            },
            OrcidRecord(
                orcid_identifier=OrcidIdentifier(
                    path="0009-0004-3041-5706",
                    uri="https://orcid.org/0009-0004-3041-5706",
                ),
                person=OrcidPerson(
                    emails=OrcidEmails(email=[OrcidEmail(email=["test@example.com"])]),
                    name=OrcidName(
                        given_names=OrcidGivenNames(value="Josiah"),
                        family_name=OrcidFamilyName(value="Carberry"),
                        visibility="private",
                    ),
                ),
            ),
        ),
    ],
    ids=["valid data", "missing email field", "visibility is private"],
)
def test_orcid_record(orcid_data: object, expected_result: OrcidRecord) -> None:
    result = OrcidRecord.model_validate(orcid_data)
    assert result == expected_result



from uuid import UUID

import pytest

from mex.common.identity import get_provider
from mex.common.ldap.extract import (
    get_merged_ids_by_employee_ids,
    get_merged_ids_by_query_string,
)
from mex.common.ldap.models import LDAPPerson, LDAPPersonWithQuery
from mex.common.types import Identifier, MergedPrimarySourceIdentifier


@pytest.fixture
def ldap_person_with_identity(
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
) -> LDAPPerson:
    person = LDAPPerson(
        objectGUID=UUID(int=1, version=4),
        employeeID="1",
        givenName=["Anna"],
        mail=["annaWithIdentity@example.com", "annaWithIdentity@test.com"],
        sn="Has Identity",
    )
    provider = get_provider()
    provider.assign(extracted_primary_source_ids["ldap"], str(person.objectGUID))
    return person


@pytest.fixture
def ldap_person_with_identity_with_query(
    ldap_person_with_identity: LDAPPerson,
) -> LDAPPersonWithQuery:
    return LDAPPersonWithQuery(person=ldap_person_with_identity, query="foo")


@pytest.fixture
def ldap_person_without_identity() -> LDAPPerson:
    return LDAPPerson(
        objectGUID=UUID(int=2, version=4),
        employeeID="2",
        givenName=["Berta"],
        mail=["bertaWithoutIdentity@example.com"],
        sn="Without Identity",
    )


@pytest.fixture
def ldap_person_without_identity_with_query(
    ldap_person_without_identity: LDAPPerson,
) -> LDAPPersonWithQuery:
    return LDAPPersonWithQuery(person=ldap_person_without_identity, query="foo")


@pytest.fixture
def ldap_persons(
    ldap_person_with_identity: LDAPPerson, ldap_person_without_identity: LDAPPerson
) -> list[LDAPPerson]:
    return [
        ldap_person_with_identity,
        ldap_person_without_identity,
    ]


@pytest.fixture
def ldap_persons_with_query(
    ldap_person_with_identity_with_query: LDAPPersonWithQuery,
    ldap_person_without_identity_with_query: LDAPPersonWithQuery,
) -> list[LDAPPersonWithQuery]:
    return [
        ldap_person_with_identity_with_query,
        ldap_person_without_identity_with_query,
    ]


@pytest.fixture
def merged_id_of_person_with_identity(
    ldap_person_with_identity: LDAPPerson,
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
) -> Identifier:
    provider = get_provider()
    identities = provider.fetch(
        had_primary_source=extracted_primary_source_ids["ldap"],
        identifier_in_primary_source=str(ldap_person_with_identity.objectGUID),
    )
    return identities[0].stableTargetId


def test_get_merged_ids_by_employee_ids(
    ldap_persons: list[LDAPPerson],
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
    merged_id_of_person_with_identity: Identifier,
    ldap_person_with_identity: LDAPPerson,
) -> None:
    expected = {
        ldap_person_with_identity.employeeID: [merged_id_of_person_with_identity]
    }
    merged_ids_by_employee_ids = get_merged_ids_by_employee_ids(
        ldap_persons, extracted_primary_source_ids["ldap"]
    )
    assert merged_ids_by_employee_ids == expected


def test_get_merged_ids_by_query_string(
    ldap_person_with_identity_with_query: LDAPPersonWithQuery,
    ldap_persons_with_query: list[LDAPPersonWithQuery],
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
    merged_id_of_person_with_identity: Identifier,
) -> None:
    expected = {
        ldap_person_with_identity_with_query.query: [merged_id_of_person_with_identity]
    }
    merged_ids_by_query_string = get_merged_ids_by_query_string(
        ldap_persons_with_query, extracted_primary_source_ids["ldap"]
    )
    assert merged_ids_by_query_string == expected



import json
from datetime import timedelta, timezone
from enum import Enum
from pathlib import Path, PurePosixPath, PureWindowsPath
from typing import Annotated, Any
from uuid import UUID

import pytest
from pydantic import AnyUrl, Field, SecretStr
from pydantic import BaseModel as PydanticModel

from mex.common.transform import (
    MExEncoder,
    dromedary_to_kebab,
    dromedary_to_snake,
    ensure_postfix,
    ensure_prefix,
    kebab_to_camel,
    normalize,
    snake_to_dromedary,
    split_to_caps,
    to_key_and_values,
)
from mex.common.types import Identifier, PathWrapper, TemporalEntity


class DummyModel(PydanticModel):
    string_field: Annotated[str, Field(alias="strField")] = "bar"
    integer: int = 42


class DummyEnum(Enum):
    THIS = "this"
    THAT = "that"


@pytest.mark.parametrize(
    ("raw", "expected"),
    [
        (DummyModel(strField="bar"), '{"integer": 42, "string_field": "bar"}'),
        (
            AnyUrl("http://example:8000/path/?query=test"),
            '"http://example:8000/path/?query=test"',
        ),
        (SecretStr("str"), '"str"'),
        (DummyEnum.THAT, '"that"'),
        (UUID(int=4, version=4), '"00000000-0000-4000-8000-000000000004"'),
        (Identifier.generate(seed=4), '"bFQoRhcVH5DHUu"'),
        (TemporalEntity(2012), '"2012"'),
        (TemporalEntity(2010, 12), '"2010-12"'),
        (TemporalEntity(2010, 12, 24), '"2010-12-24"'),
        (TemporalEntity(2010, 12, 24, 23, 59, 59), '"2010-12-24T23:41:59Z"'),
        (
            TemporalEntity(
                2010, 12, 24, 23, 59, 59, tzinfo=timezone(timedelta(hours=-2))
            ),
            '"2010-12-25T01:59:59Z"',
        ),
        (PureWindowsPath(r"C:\\System\\Win32\\exe.dll"), '"C:/System/Win32/exe.dll"'),
        (PurePosixPath(r"/dev/sys/etc/launch.ctl"), '"/dev/sys/etc/launch.ctl"'),
        (Path("relative", "path"), '"relative/path"'),
        (PathWrapper("relative/path"), '"relative/path"'),
    ],
)
def test_mex_json_encoder(raw: object, expected: str) -> None:
    assert json.dumps(raw, cls=MExEncoder, sort_keys=True) == expected


def test_mex_json_encoder_unserializable() -> None:
    with pytest.raises(TypeError, match="object is not JSON serializable"):
        assert json.dumps({"foo": object()}, cls=MExEncoder)


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", ""),
        ("word", "word"),
        ("alreadyDromedary", "alreadyDromedary"),
        ("multiple_words_in_a_string", "multipleWordsInAString"),
    ],
    ids=[
        "empty",
        "single word",
        "already dromedary",
        "multiple words",
    ],
)
def test_snake_to_dromedary(string: str, expected: str) -> None:
    result = snake_to_dromedary(string)
    assert result == expected


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", ""),
        ("word", "word"),
        ("already_snake", "already_snake"),
        ("ABWordCDEWordFG", "ab_word_cde_word_fg"),
        ("multipleWordsInAString", "multiple_words_in_a_string"),
    ],
    ids=[
        "empty",
        "single word",
        "already snake",
        "caps words",
        "multiple words",
    ],
)
def test_dromedary_to_snake(string: str, expected: str) -> None:
    result = dromedary_to_snake(string)
    assert result == expected


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", ""),
        ("word", "word"),
        ("already-kebab", "already-kebab"),
        ("Such-AWeird-MIXEDCase", "such-a-weird-mixed-case"),
        ("ABWordCDEWordFG", "ab-word-cde-word-fg"),
        ("multipleWordsInAString", "multiple-words-in-a-string"),
    ],
    ids=[
        "empty",
        "single word",
        "already kebab",
        "mixed case",
        "caps words",
        "multiple words",
    ],
)
def test_dromedary_to_kebab(string: str, expected: str) -> None:
    result = dromedary_to_kebab(string)
    assert result == expected


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", ""),
        ("word", "Word"),
        ("AlreadyCamel", "AlreadyCamel"),
        ("Mixed-CASE", "MixedCase"),
        ("multiple-words-in-a-string", "MultipleWordsInAString"),
    ],
    ids=[
        "empty",
        "single word",
        "already camel",
        "mixed case",
        "multiple words",
    ],
)
def test_kebab_to_camel(string: str, expected: str) -> None:
    result = kebab_to_camel(string)
    assert result == expected


@pytest.mark.parametrize(
    ("string", "expected"),
    [("", ""), ("__XYZ__", "xyz"), ("/foo/BAR$42", "foo bar 42")],
)
def test_normalize(string: str, expected: str) -> None:
    assert normalize(string) == expected


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("", ""),
        ("Foo(Bar) 99 - Batz", "FOO_BAR_BATZ"),
    ],
)
def test_split_to_caps(string: str, expected: str) -> None:
    assert split_to_caps(string) == expected


@pytest.mark.parametrize(
    ("string", "prefix", "expected"),
    [
        ("", "", ""),
        ("banana", "ba", "banana"),
        ("bar", "foo", "foobar"),
        (
            -42,
            UUID("{12345678-1234-5678-1234-567812345678}"),
            "12345678-1234-5678-1234-567812345678-42",
        ),
    ],
    ids=["empty", "already-prefixed", "prefix-added", "stringified"],
)
def test_ensure_prefix(string: object, prefix: object, expected: str) -> None:
    result = ensure_prefix(string, prefix)

    assert result == expected


@pytest.mark.parametrize(
    ("string", "postfix", "expected"),
    [
        ("", "", ""),
        ("banana", "na", "banana"),
        ("foo", "bar", "foobar"),
        (
            UUID("{12345678-1234-5678-1234-567812345678}"),
            -42,
            "12345678-1234-5678-1234-567812345678-42",
        ),
    ],
    ids=["empty", "already-postfixed", "postfix-added", "stringified"],
)
def test_ensure_postfix(
    string: object,
    postfix: object,
    expected: str,
) -> None:
    result = ensure_postfix(string, postfix)

    assert result == expected


@pytest.mark.parametrize(
    ("dct", "expected"),
    [
        ({}, {}),
        (
            {"single": 32, "nested": {"foo": 42}, "empty": None},
            {"single": [32], "nested": [{"foo": 42}], "empty": []},
        ),
        (
            {"one": [32], "three": [32, 42, 3.1], "empty": []},
            {"one": [32], "three": [32, 42, 3.1], "empty": []},
        ),
    ],
    ids=["empty", "singles", "lists"],
)
def test_to_key_and_values(dct: dict[str, Any], expected: dict[str, list[Any]]) -> None:
    result = dict(to_key_and_values(dct))

    assert result == expected



from typing import get_origin

from pydantic_core import PydanticUndefined

from mex.common.models import (
    EXTRACTED_MODEL_CLASSES,
    EXTRACTED_MODEL_CLASSES_BY_NAME,
    MAPPING_MODEL_CLASSES,
    MappingField,
    VariableGroupMapping,
)


def test_all_mapping_classes_are_defined() -> None:
    stem_types = sorted(c.stemType for c in EXTRACTED_MODEL_CLASSES)
    assert sorted(c.stemType for c in MAPPING_MODEL_CLASSES) == stem_types


def test_all_mapping_fields_are_defined() -> None:
    for mapping_cls in MAPPING_MODEL_CLASSES:
        extracted_cls = EXTRACTED_MODEL_CLASSES_BY_NAME[
            f"Extracted{mapping_cls.stemType}"
        ]
        assert set(mapping_cls.model_fields) == set(extracted_cls.model_fields)
        field_defs = {
            field_name: (field_info.annotation, field_info.default)
            for field_name, field_info in mapping_cls.model_fields.items()
            if field_name != "entityType"
        }
        assert all(
            annotation is not None
            and get_origin(annotation) is list
            and annotation.__args__[0].__bases__[0] is MappingField
            and default in (PydanticUndefined, [])
            for annotation, default in field_defs.values()
        )


def test_mapping_model_schema() -> None:
    assert VariableGroupMapping.model_json_schema() == {
        "$defs": {
            "MappingField_MergedPrimarySourceIdentifier_": {
                "additionalProperties": False,
                "properties": {
                    "fieldInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "fieldInPrimarySource",
                    },
                    "locationInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "locationInPrimarySource",
                    },
                    "examplesInPrimarySource": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "examplesInPrimarySource",
                    },
                    "mappingRules": {
                        "items": {
                            "$ref": "#/$defs/MappingRule_MergedPrimarySourceIdentifier_"
                        },
                        "minItems": 1,
                        "title": "mappingRules",
                        "type": "array",
                    },
                    "comment": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "comment",
                    },
                },
                "required": ["mappingRules"],
                "title": "MappingField[MergedPrimarySourceIdentifier]",
                "type": "object",
            },
            "MappingField_list_MergedResourceIdentifier__": {
                "additionalProperties": False,
                "properties": {
                    "fieldInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "fieldInPrimarySource",
                    },
                    "locationInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "locationInPrimarySource",
                    },
                    "examplesInPrimarySource": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "examplesInPrimarySource",
                    },
                    "mappingRules": {
                        "items": {
                            "$ref": "#/$defs/MappingRule_list_MergedResourceIdentifier__"
                        },
                        "minItems": 1,
                        "title": "mappingRules",
                        "type": "array",
                    },
                    "comment": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "comment",
                    },
                },
                "required": ["mappingRules"],
                "title": "MappingField[list[MergedResourceIdentifier]]",
                "type": "object",
            },
            "MappingField_list_Text__": {
                "additionalProperties": False,
                "properties": {
                    "fieldInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "fieldInPrimarySource",
                    },
                    "locationInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "locationInPrimarySource",
                    },
                    "examplesInPrimarySource": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "examplesInPrimarySource",
                    },
                    "mappingRules": {
                        "items": {"$ref": "#/$defs/MappingRule_list_Text__"},
                        "minItems": 1,
                        "title": "mappingRules",
                        "type": "array",
                    },
                    "comment": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "comment",
                    },
                },
                "required": ["mappingRules"],
                "title": "MappingField[list[Text]]",
                "type": "object",
            },
            "MappingField_str_": {
                "additionalProperties": False,
                "properties": {
                    "fieldInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "fieldInPrimarySource",
                    },
                    "locationInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "locationInPrimarySource",
                    },
                    "examplesInPrimarySource": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "examplesInPrimarySource",
                    },
                    "mappingRules": {
                        "items": {"$ref": "#/$defs/MappingRule_str_"},
                        "minItems": 1,
                        "title": "mappingRules",
                        "type": "array",
                    },
                    "comment": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "comment",
                    },
                },
                "required": ["mappingRules"],
                "title": "MappingField[str]",
                "type": "object",
            },
            "MappingRule_MergedPrimarySourceIdentifier_": {
                "additionalProperties": False,
                "properties": {
                    "forValues": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "forValues",
                    },
                    "setValues": {
                        "anyOf": [
                            {
                                "pattern": "^[a-zA-Z0-9]{14,22}$",
                                "title": "MergedPrimarySourceIdentifier",
                                "type": "string",
                            },
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "setValues",
                    },
                    "rule": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "rule",
                    },
                },
                "title": "MappingRule[MergedPrimarySourceIdentifier]",
                "type": "object",
            },
            "MappingRule_list_MergedResourceIdentifier__": {
                "additionalProperties": False,
                "properties": {
                    "forValues": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "forValues",
                    },
                    "setValues": {
                        "anyOf": [
                            {
                                "items": {
                                    "pattern": "^[a-zA-Z0-9]{14,22}$",
                                    "title": "MergedResourceIdentifier",
                                    "type": "string",
                                },
                                "type": "array",
                            },
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "setValues",
                    },
                    "rule": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "rule",
                    },
                },
                "title": "MappingRule[list[MergedResourceIdentifier]]",
                "type": "object",
            },
            "MappingRule_list_Text__": {
                "additionalProperties": False,
                "properties": {
                    "forValues": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "forValues",
                    },
                    "setValues": {
                        "anyOf": [
                            {"items": {"$ref": "#/$defs/Text"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "setValues",
                    },
                    "rule": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "rule",
                    },
                },
                "title": "MappingRule[list[Text]]",
                "type": "object",
            },
            "MappingRule_str_": {
                "additionalProperties": False,
                "properties": {
                    "forValues": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "forValues",
                    },
                    "setValues": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "setValues",
                    },
                    "rule": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "rule",
                    },
                },
                "title": "MappingRule[str]",
                "type": "object",
            },
            "Text": {
                "description": 'Type class for text objects.\n\nTexts can be parsed from nested JSON objects or from raw strings.\n\nExample:\n    Text(value="foo") == Text.model_validate("foo")',
                "properties": {
                    "value": {"minLength": 1, "title": "Value", "type": "string"},
                    "language": {
                        "anyOf": [{"$ref": "#/$defs/TextLanguage"}, {"type": "null"}],
                        "default": None,
                    },
                },
                "required": ["value"],
                "title": "Text",
                "type": "object",
            },
            "TextLanguage": {
                "description": "Possible language tags for `Text` values.",
                "enum": ["de", "en", "fr", "es", "ru"],
                "title": "TextLanguage",
                "type": "string",
            },
        },
        "additionalProperties": False,
        "description": "Mapping for describing a variable group transformation.",
        "properties": {
            "$type": {
                "const": "VariableGroupMapping",
                "default": "VariableGroupMapping",
                "title": "$Type",
                "type": "string",
            },
            "hadPrimarySource": {
                "items": {
                    "$ref": "#/$defs/MappingField_MergedPrimarySourceIdentifier_"
                },
                "minItems": 1,
                "title": "Hadprimarysource",
                "type": "array",
            },
            "identifierInPrimarySource": {
                "items": {"$ref": "#/$defs/MappingField_str_"},
                "minItems": 1,
                "title": "Identifierinprimarysource",
                "type": "array",
            },
            "containedBy": {
                "items": {
                    "$ref": "#/$defs/MappingField_list_MergedResourceIdentifier__"
                },
                "minItems": 1,
                "title": "Containedby",
                "type": "array",
            },
            "label": {
                "items": {"$ref": "#/$defs/MappingField_list_Text__"},
                "minItems": 1,
                "title": "Label",
                "type": "array",
            },
        },
        "required": [
            "hadPrimarySource",
            "identifierInPrimarySource",
            "containedBy",
            "label",
        ],
        "title": "VariableGroupMapping",
        "type": "object",
    }



from dataclasses import dataclass

import pytest

from mex.common.exceptions import MExError
from mex.common.sorters import topological_sort


@dataclass
class Item:
    id: str
    parent: list[str] | str | None = None
    child: list[str] | str | None = None


@pytest.mark.parametrize(
    ("items", "expected"),
    [
        (
            [
                Item(id="ChildB", parent="Parent"),
                Item(id="GrandParent"),
                Item(id="ChildA", parent="Parent"),
                Item(id="Parent", parent="GrandParent"),
            ],
            [
                Item(id="GrandParent"),
                Item(id="Parent", parent="GrandParent"),
                Item(id="ChildB", parent="Parent"),
                Item(id="ChildA", parent="Parent"),
            ],
        ),
        (
            [
                Item(id="ChildB", parent=["ParentA", "ParentB"]),
                Item(id="ParentB"),
                Item(id="ChildC", parent="ParentA"),
                Item(id="ParentA"),
                Item(id="ChildA", parent=["ParentB", "ParentA"]),
            ],
            [
                Item(id="ParentA"),
                Item(id="ParentB"),
                Item(id="ChildC", parent="ParentA"),
                Item(id="ChildB", parent=["ParentA", "ParentB"]),
                Item(id="ChildA", parent=["ParentB", "ParentA"]),
            ],
        ),
        (
            [
                Item(id="ChildB"),
                Item(id="ParentB", child=["ChildB"]),
                Item(id="ChildC", parent="ParentA"),
                Item(id="ParentA", child=["ChildB", "ChildA"]),
                Item(id="ChildA", parent=["ParentB", "ParentA"]),
            ],
            [
                Item(id="ParentB", child=["ChildB"]),
                Item(id="ParentA", child=["ChildB", "ChildA"]),
                Item(id="ChildC", parent="ParentA"),
                Item(id="ChildB"),
                Item(id="ChildA", parent=["ParentB", "ParentA"]),
            ],
        ),
    ],
    ids=["single parent", "multi parents", "two way"],
)
def test_topological_sort(items: list[Item], expected: list[Item]) -> None:
    topological_sort(items, "id", parent_key="parent", child_key="child")
    assert items == expected


def test_topological_sort_circular_error() -> None:
    items = [
        Item(id="North", child=["East"]),
        Item(id="East", child=["South"]),
        Item(id="South", child=["West"]),
        Item(id="West", child=["North"]),
    ]
    with pytest.raises(MExError, match=r"Found graph cycles while sorting"):
        topological_sort(items, "id", child_key="child")



import json
import pdb  # noqa: T100
import sys
from bdb import BdbQuit
from collections.abc import Callable
from functools import partial
from textwrap import dedent
from traceback import format_exc
from typing import Any

import click
from click import Command, Option
from click.core import ParameterSource
from click.exceptions import Abort, Exit
from pydantic.fields import FieldInfo

from mex.common.connector import CONNECTOR_STORE
from mex.common.logging import logger
from mex.common.settings import SETTINGS_STORE, BaseSettings
from mex.common.transform import MExEncoder

HELP_TEMPLATE = """
{doc}

Acceptable configuration sources sorted by priority:
(1) command line arguments and options
(2) environment variables
(3) dotenv file located at {env_file}
(4) default values from settings model
"""


def _field_to_parameters(name: str, field: FieldInfo) -> list[str]:
    """Convert a field of a pydantic settings class into parameter declarations.

    The field's name and alias are considered. Underscores are replaced with dashes
    and single character parameters have two leading dashes while single character
    parameters have just one.

    Args:
        name: name of the Field
        field: Field of a Settings definition class

    Returns:
        List of parameter declaring strings
    """
    names = [name] + ([field.alias] if field.alias else [])
    names = [n.replace("_", "-") for n in names]
    dashes = ["--" if len(n) > 1 else "-" for n in names]
    return [f"{d}{n}" for d, n in zip(dashes, names, strict=False)]


def _field_to_option(name: str, settings_cls: type[BaseSettings]) -> Option:
    """Convert a field of a pydantic settings class into a click option.

    Args:
        name: name of the Field
        settings_cls: Base settings class or a subclass of it

    Returns:
        Option: click Option with appropriate attributes
    """
    # normalize field type to be compatible with advanced string types
    # https://pydantic-docs.helpmanual.io/usage/types/#pydantic-types
    # complex fields or type unions are always interpreted as strings
    # and add support for SecretStr fields with correct default values
    # https://pydantic-docs.helpmanual.io/usage/types/#secret-types
    field = settings_cls.model_fields[name]

    if field.annotation in (int, bool, float):
        field_type: Any = field.annotation
    else:
        field_type = str

    if field.is_required():
        default = None
    elif field.annotation in (int, bool, float):
        default = field.default
    else:
        default = json.dumps(field.default, cls=MExEncoder).strip('"')

    return Option(
        _field_to_parameters(name, field),
        default=default,
        envvar=settings_cls.get_env_name(name),
        help=field.description,
        is_flag=field.annotation is bool and field.default is False,
        show_default=True,
        show_envvar=True,
        type=field_type,
        required=field.is_required(),
    )


def _callback(
    func: Callable[[], None],
    settings_cls: type[BaseSettings],
    **cli_settings: str,
) -> None:
    """Run the decorated function in the current click context.

    When `cli_settings` specify debug mode and an exception occurs,
    jump into post mortem debugging and raise exception.

    Args:
        func: Entry point function for a cli
        settings_cls: Base settings class or a subclass of it
        cli_settings: Parsed settings in string format

    Raises:
        Exception: Any uncaught exception when in debug mode
        SysExit: With exit code 0 or 1
    """
    # get current click context.
    context = click.get_current_context()

    # ensure all singletons are reset.
    context.call_on_close(CONNECTOR_STORE.reset)
    context.call_on_close(SETTINGS_STORE.reset)

    # load settings from parameters and store it globally.
    settings = settings_cls.model_validate(
        {
            key: value
            for key, value in cli_settings.items()
            if context.get_parameter_source(key) == ParameterSource.COMMANDLINE
        }
    )
    SETTINGS_STORE.push(settings)

    # otherwise print loaded settings in pretty way and continue.
    logger.info(click.style(dedent(f"    {func.__doc__}"), fg="green"))
    logger.info(click.style(f"{settings.text()}\n", fg="bright_cyan"))

    # now try to execute the decorated function.
    try:
        func()
    except (Abort, BdbQuit, Exit, KeyboardInterrupt):  # pragma: no cover
        context.exit(130)
    except Exception:
        # an error occurred, let's print the traceback
        logger.error(click.style(format_exc(), fg="red"))
        if settings.debug:  # pragma: no cover
            # if we are in debug mode, jump into interactive debugging.
            pdb.post_mortem(sys.exc_info()[2])
            raise
        # if not in debug mode, exit with code 1.
        logger.error("exit")
        context.exit(1)

    # all good, exit with code 0.
    logger.info("done")
    context.exit(0)


def entrypoint(
    settings_cls: type[BaseSettings],
) -> Callable[[Callable[[], None]], Command]:
    """Decorate given function to mark it as a cli entrypoint.

    The decorator takes one argument `settings_cls` that is either
    `mex.common.settings.BaseSettings` or a subclass thereof. The decorated function
    must not require any positional arguments and does not need to return anything.

    Running an `entrypoint` will print a summary on startup, register settings and
    connector singletons globally and provide error handling as well as debugging.

    Args:
        settings_cls: Settings class that should be instantiated globally.

    Returns:
        Callable: The decorated function with initialized settings.
    """

    def decorator(func: Callable[[], None]) -> Command:
        return Command(
            func.__name__,
            help=HELP_TEMPLATE.format(
                doc=func.__doc__, env_file=settings_cls.model_config.get("env_file")
            ),
            callback=partial(_callback, func, settings_cls),
            params=[
                _field_to_option(name, settings_cls)
                for name in settings_cls.model_fields
            ],
        )

    return decorator



from mex.common.connector.base import CONNECTOR_STORE, BaseConnector
from mex.common.connector.http import HTTPConnector

__all__ = (
    "CONNECTOR_STORE",
    "BaseConnector",
    "HTTPConnector",
)



"""Helper extractor to extract data from Lightweight Directory Access Protocol (LDAP).

Common use cases:
- extract employee accounts of your organization
- extract functional accounts of your organization

Possible queries are for example the account name, surname, given name, or email.

Configuration
-------------

For configuring the ldap connection, set the settings parameter `ldap_url`
(see `mex.common.settings` for further info) to an LDAP url (see
 https://datatracker.ietf.org/doc/html/rfc2255#section-3 for further information).

Extracting data
---------------

Use the `LDAPConnector` from the `ldap.connector` module to extract data.

Transforming data
-----------------

The module `ldap.transform` contains functions for transforming LDAP data into MEx
models.

The `mex_person.stableTargetId` attribute can be used in any entity that requires a
`MergedPersonIdentifier`.

Convenience Functions
---------------------

The module `ldap.extract` holds convenience functions, e.g. for build a mapping from
query strings to `stableTargetId`s.

"""



import json
import re
from collections.abc import Callable
from copy import deepcopy
from itertools import zip_longest
from typing import Any

import pytest

from mex.common.models import EXTRACTED_MODEL_CLASSES, BaseModel
from mex.common.transform import dromedary_to_kebab
from mex.common.types import IDENTIFIER_PATTERN, VOCABULARY_PATTERN
from mex.model import ENTITY_JSON_BY_NAME


def model_to_schema(model: type[BaseModel]) -> dict[str, Any]:
    # pydantic does not include computed fields in the validation schema
    # and does not include validation rules in the serialization schema.
    # so we need to mangle those two together here, to get a schema that is
    # more comparable to what mex-model specifies.

    validation_schema = model.model_json_schema(
        ref_template="/schema/fields/{model}", mode="validation"
    )
    serialization_schema = model.model_json_schema(
        ref_template="/schema/fields/{model}", mode="serialization"
    )
    validation_schema["properties"] = {
        **serialization_schema["properties"],
        **validation_schema["properties"],
    }
    validation_schema["required"] = sorted(
        {*serialization_schema["required"], *validation_schema["required"]}
    )
    return validation_schema


GENERATED_SCHEMAS = dict(
    sorted(
        {
            schema["title"]: schema
            for schema in [model_to_schema(model) for model in EXTRACTED_MODEL_CLASSES]
        }.items()
    )
)
SPECIFIED_SCHEMAS = dict(
    sorted({schema["title"]: schema for schema in ENTITY_JSON_BY_NAME.values()}.items())
)
ENTITY_TYPES_AND_FIELD_NAMES_BY_FQN = {
    f"{entity_type}.{field_name}": (entity_type, field_name)
    for entity_type, schema in SPECIFIED_SCHEMAS.items()
    for field_name in schema["properties"]
}


def test_entity_types_match_spec() -> None:
    assert list(GENERATED_SCHEMAS) == list(SPECIFIED_SCHEMAS)


@pytest.mark.parametrize(
    ("generated", "specified"),
    zip_longest(GENERATED_SCHEMAS.values(), SPECIFIED_SCHEMAS.values(), fillvalue={}),
    ids=map(str, zip_longest(GENERATED_SCHEMAS, SPECIFIED_SCHEMAS, fillvalue="N/A")),
)
def test_field_names_match_spec(
    generated: dict[str, Any], specified: dict[str, Any]
) -> None:
    ignored_field_names = {
        "$type",  # only in generated
    }
    generated_field_names = {
        k: v for k, v in generated["properties"].items() if k not in ignored_field_names
    }
    specified_field_names = set(specified["properties"])
    assert set(generated_field_names) == specified_field_names


@pytest.mark.parametrize(
    ("generated", "specified"),
    zip_longest(GENERATED_SCHEMAS.values(), SPECIFIED_SCHEMAS.values(), fillvalue={}),
    ids=map(str, zip_longest(GENERATED_SCHEMAS, SPECIFIED_SCHEMAS, fillvalue="N/A")),
)
def test_entity_type_matches_metadata(
    generated: dict[str, Any], specified: dict[str, Any]
) -> None:
    ignored_class_meta = {
        *("$$target", "$comment", "$defs", "$id", "$schema"),  # skip "$" fields
        "properties",  # checked in other tests
    }
    generated_meta = {k: v for k, v in generated.items() if k not in ignored_class_meta}
    specified_meta = {k: v for k, v in specified.items() if k not in ignored_class_meta}
    assert generated_meta == specified_meta


def dissolve_single_item_lists(dct: dict[str, Any], key: str) -> None:
    # if a list in a dict value has just one item, dissolve it into the parent dict
    if len(dct[key]) == 1 and isinstance(dct[key][0], dict):
        dct.update(dct.pop(key)[0])


def sub_only_text(repl: Callable[[str], str], string: str) -> str:
    # substitute only the textual parts of a string, e.g. leave slashes alone
    return re.sub(r"([a-zA-Z_-]+)", lambda m: repl(m.group(0)), string)


def prepare_field(field: str, obj: list[Any] | dict[str, Any]) -> None:
    # prepare each item in a list (in-place)
    if isinstance(obj, list):
        for item in obj:
            prepare_field(field, item)
        obj[:] = [item for item in obj if item]
        return

    # discard comments and descriptions
    obj.pop("$comment", None)  # only in spec
    obj.pop("description", None)  # only in spec

    # discard title but save value for later
    title = obj.pop("title", field)  # only in generated

    # align reference paths
    # (the paths to referenced vocabularies and types differ between the models
    # and the specification, so we need to make sure they match before comparing)
    if obj.get("pattern") == IDENTIFIER_PATTERN:
        del obj["pattern"]
        del obj["type"]
        if field in ("identifier", "stableTargetId"):
            obj["$ref"] = "/schema/fields/identifier"
        else:
            obj["$ref"] = "/schema/entities/{}#/identifier".format(
                title.removesuffix("Identifier").removeprefix("Merged")
            )

    # align concept/enum annotations
    elif obj.get("$ref") == "/schema/entities/concept#/identifier":
        obj["pattern"] = VOCABULARY_PATTERN
        obj["type"] = "string"
        del obj["$ref"]

    # make sure all refs have paths in kebab-case
    # (the models use the class names, whereas the spec uses kebab-case URLs)
    if "$ref" in obj:
        obj["$ref"] = sub_only_text(dromedary_to_kebab, obj["$ref"])

    # recurse into the field definitions for array items
    if obj.get("type") == "array":
        prepare_field(field, obj["items"])

    for quantifier in {"anyOf", "allOf"} & set(obj):
        # prepare choices
        prepare_field(field, obj[quantifier])
        # collapse non-choices
        dissolve_single_item_lists(obj, quantifier)


@pytest.mark.parametrize(
    ("entity_type", "field_name"),
    ENTITY_TYPES_AND_FIELD_NAMES_BY_FQN.values(),
    ids=ENTITY_TYPES_AND_FIELD_NAMES_BY_FQN.keys(),
)
def test_field_defs_match_spec(entity_type: str, field_name: str) -> None:
    specified_properties = SPECIFIED_SCHEMAS[entity_type]["properties"]
    generated_properties = GENERATED_SCHEMAS[entity_type]["properties"]
    specified = deepcopy(specified_properties[field_name])
    generated = deepcopy(generated_properties[field_name])

    prepare_field(field_name, specified)
    prepare_field(field_name, generated)

    assert generated == specified, f"""
{entity_type}.{field_name}

specified:
{json.dumps(specified_properties[field_name], indent=4, sort_keys=True)}

generated:
{json.dumps(generated_properties[field_name], indent=4, sort_keys=True)}
"""



from collections import defaultdict
from collections.abc import Generator
from os import PathLike
from typing import TYPE_CHECKING, Any, TypeVar, Union

import numpy as np
import pandas as pd
from pydantic import ValidationError

from mex.common.logging import logger

if TYPE_CHECKING:  # pragma: no cover
    from pandas._typing import Dtype, ReadCsvBuffer

    from mex.common.models import BaseModel

_BaseModelT = TypeVar("_BaseModelT", bound="BaseModel")

PANDAS_DTYPE_MAP = defaultdict(
    lambda: "string",
    {bool: "bool", float: "Float64", int: "Int64"},
)


def get_dtypes_for_model(model: type["BaseModel"]) -> dict[str, "Dtype"]:
    """Get the basic dtypes per field for a model from the `PANDAS_DTYPE_MAP`.

    Args:
        model: Model class for which to get pandas data types per field alias

    Returns:
        Mapping from field alias to dtype strings
    """
    return {
        f.alias or name: PANDAS_DTYPE_MAP[f.annotation or type(None)]
        for name, f in model.model_fields.items()
    }


def parse_csv(
    path_or_buffer: Union[str, PathLike[str], "ReadCsvBuffer[Any]"],
    into: type[_BaseModelT],
    chunksize: int = 10000,
    summary_batch_size: int = 10000,
    **kwargs: Any,  # noqa: ANN401
) -> Generator[_BaseModelT, None, None]:
    """Parse a CSV file into an iterable of the given model type.

    Args:
        path_or_buffer: Location of CSV file or read buffer with CSV content
        into: Type of model to parse
        chunksize: Buffer size for chunked reading
        summary_batch_size: Batch size for summary logs
        kwargs: Additional keywords arguments for pandas

    Returns:
        Generator for models
    """
    error_summary: defaultdict[str, int] = defaultdict(int)
    total_rows_processed = 0
    total_rows_successfully_processed = 0

    with pd.read_csv(
        path_or_buffer,
        chunksize=chunksize,
        dtype=get_dtypes_for_model(into),  # type: ignore[arg-type]
        **kwargs,
    ) as reader:
        for i, chunk in enumerate(reader):
            logger.info(
                "parse_csv - %s chunk %s - OK",
                into.__name__,
                i,
            )
            for _, row in chunk.iterrows():
                try:
                    model = into.model_validate(
                        row.replace(to_replace=np.nan, value=None)
                        .replace(regex=r"^\s*$", value=None)
                        .to_dict()
                    )
                    total_rows_successfully_processed += 1
                    yield model
                except ValidationError as error:
                    for validation_error in error.errors():
                        error_type = validation_error["type"]
                        error_summary[error_type] += 1

                total_rows_processed += 1

                if total_rows_processed % summary_batch_size == 0 and error_summary:
                    logger.error(
                        "Summarizing errors for batch with rows %s to %s",
                        total_rows_processed - summary_batch_size + 1,
                        total_rows_processed,
                    )
                    for error_type, count in error_summary.items():
                        logger.error(
                            " - Error type '%s': %s occurrences", error_type, count
                        )
                    error_summary.clear()

    if error_summary:
        logger.error("Summarizing errors for remaining rows")
        for error_type, count in error_summary.items():
            logger.error(" - Error type '%s': %s occurrences", error_type, count)
        logger.info(
            "Successfully processed %s items.", total_rows_successfully_processed
        )



from collections import defaultdict
from collections.abc import Iterable

from mex.common.identity import get_provider
from mex.common.ldap.models import LDAPPerson, LDAPPersonWithQuery
from mex.common.types import MergedPersonIdentifier, MergedPrimarySourceIdentifier


def get_merged_ids_by_employee_ids(
    persons: Iterable[LDAPPerson],
    primary_source_id: MergedPrimarySourceIdentifier,
) -> dict[str, list[MergedPersonIdentifier]]:
    """Return a mapping from a person's employeeID to their merged person ids.

    MergedPersonIdentifiers are looked up in the identity provider and will be omitted
    for any person that has not yet been assigned an `Identity` there.

    Args:
        persons: Iterable of LDAP persons
        primary_source_id: Primary source identifier for LDAP

    Returns:
        Mapping from `LDAPPerson.employeeID` to corresponding MergedPersonIdentifiers
    """
    merged_ids_by_attribute = defaultdict(list)
    provider = get_provider()
    for person in persons:
        if identities := provider.fetch(
            had_primary_source=primary_source_id,
            identifier_in_primary_source=str(person.objectGUID),
        ):
            merged_id = MergedPersonIdentifier(identities[0].stableTargetId)
            merged_ids_by_attribute[person.employeeID].append(merged_id)
    return merged_ids_by_attribute


def get_merged_ids_by_query_string(
    persons_with_query: Iterable[LDAPPersonWithQuery],
    primary_source_id: MergedPrimarySourceIdentifier,
) -> dict[str, list[MergedPersonIdentifier]]:
    """Return a mapping from a person query string to their merged person ids.

    MergedPersonIdentifiers are looked up in the identity provider and will be omitted
    for any person that has not yet been assigned an `Identity` there.

    Args:
        persons_with_query: Iterable of LDP persons with query
        primary_source_id: Primary source for LDAP

    Returns:
        Mapping from `LDAPPersonWithQuery.query` to corresponding
        MergedPersonIdentifiers
    """
    merged_ids_by_attribute = defaultdict(list)
    provider = get_provider()
    for person_with_query in persons_with_query:
        if identities := provider.fetch(
            had_primary_source=primary_source_id,
            identifier_in_primary_source=str(person_with_query.person.objectGUID),
        ):
            merged_ids_by_attribute[str(person_with_query.query)].append(
                MergedPersonIdentifier(identities[0].stableTargetId)
            )
    return merged_ids_by_attribute



from mex.common.models import ExtractedOrganization
from mex.common.organigram.models import OrganigramUnit
from mex.common.organigram.transform import (
    transform_organigram_units_to_organizational_units,
)
from mex.common.testing import Joker
from mex.common.types import (
    LinkLanguage,
    MergedPrimarySourceIdentifier,
    Text,
    TextLanguage,
)


def test_transform_organigram_units_to_organizational_units(
    child_unit: OrganigramUnit,
    parent_unit: OrganigramUnit,
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
    rki_organization: ExtractedOrganization,
) -> None:
    extracted_units = transform_organigram_units_to_organizational_units(
        [child_unit, parent_unit],
        extracted_primary_source_ids["organigram"],
        rki_organization,
    )

    # look up by ids, because order is not guaranteed
    extracted_units_by_id = {u.identifierInPrimarySource: u for u in extracted_units}
    assert len(extracted_units_by_id) == 2
    child_extracted_unit = extracted_units_by_id["child-unit"]
    parent_extracted_unit = extracted_units_by_id["parent-unit"]

    # check parent relation is resolved
    assert parent_extracted_unit.parentUnit is None
    assert child_extracted_unit.parentUnit is not None
    assert child_extracted_unit.parentUnit == parent_extracted_unit.stableTargetId

    # check name languages are set
    assert child_extracted_unit.name == [
        Text(value="CHLD Unterabteilung", language=TextLanguage.DE),
        Text(value="C1: Sub Unit", language=TextLanguage.EN),
    ]

    # check serialized as expected
    assert parent_extracted_unit.model_dump(exclude_none=True) == {
        "identifier": Joker(),
        "hadPrimarySource": extracted_primary_source_ids["organigram"],
        "identifierInPrimarySource": "parent-unit",
        "stableTargetId": Joker(),
        "alternativeName": [
            {"value": "PRNT Abteilung", "language": TextLanguage.DE},
            {"value": "PARENT Dept."},
        ],
        "email": ["pu@example.com", "PARENT@example.com"],
        "name": [
            {"value": "Abteilung", "language": TextLanguage.DE},
            {"value": "Department", "language": TextLanguage.EN},
        ],
        "shortName": [{"value": "PRNT"}],
        "unitOf": ["hDGPcPIlVrQZu9EbokVFmm"],
        "website": [
            {
                "language": LinkLanguage.EN,
                "title": "Example | Parent Department",
                "url": "https://www.example.com/departments/parent.html",
            }
        ],
        "entityType": "ExtractedOrganizationalUnit",
    }



from datetime import date, datetime
from typing import Any

import pytest
from pydantic import BaseModel
from pytz import timezone

from mex.common.types import (
    UTC,
    TemporalEntity,
    TemporalEntityPrecision,
    Year,
    YearMonth,
    YearMonthDay,
    YearMonthDayTime,
)
from mex.common.types.temporal_entity import YEAR_MONTH_DAY_REGEX

CET_tz = timezone("Europe/Berlin")
PST_tz = timezone("America/Los_Angeles")


@pytest.mark.parametrize(
    ("args", "kwargs", "message"),
    [
        (
            (datetime.now(tz=UTC),),
            {"tzinfo": UTC},
            "Temporal entity does not accept tzinfo in parsing mode",
        ),
        ((0,) * 10, {}, "Temporal entity takes at most 7 arguments"),
        (
            (1.3, "foo", 42),
            {},
            "Temporal entity takes a single str, date, datetime or TemporalEntity "
            "argument or up to 7 integers",
        ),
    ],
)
def test_temporal_entity_type_errors(
    args: tuple[Any], kwargs: dict[str, Any], message: str
) -> None:
    with pytest.raises(TypeError, match=message):
        TemporalEntity(*args, **kwargs)


@pytest.mark.parametrize(
    ("cls", "args", "kwargs", "error"),
    [
        (
            YearMonthDayTime,
            ("2001-04-24",),
            {},
            "Expected precision level to be one of 'hour', 'minute', 'second', 'microsecond'",
        ),
        (
            YearMonthDay,
            ("1999-02",),
            {},
            "Expected precision level to be 'day'",
        ),
        (
            YearMonth,
            ("1999-02",),
            {"precision": TemporalEntityPrecision.DAY},
            "Expected precision level to be 'month'",
        ),
        (
            YearMonthDay,
            ("2022-02-31",),
            {},
            "day is out of range for month",
        ),
        (
            YearMonthDay,
            ("2023-02-29",),
            {},
            "day is out of range for month",
        ),
        (
            YearMonth,
            ("1987-13",),
            {},
            "month must be in 1..12",
        ),
        (
            YearMonthDayTime,
            ("2018-03-02T12:61:01Z",),
            {},
            "minute must be in 0..59",
        ),
    ],
    ids=[
        "too precise",
        "not precise enough",
        "wrong explicit precision",
        "not a valid day in any year",
        "not a valid leap day",
        "not a valid month",
        "not a valid minute",
    ],
)
def test_temporal_entity_value_errors(
    cls: type[TemporalEntity], args: tuple[Any], kwargs: dict[str, Any], error: str
) -> None:
    with pytest.raises(ValueError, match=error):
        cls(*args, **kwargs)


@pytest.mark.parametrize(
    ("cls", "args", "kwargs", "expected"),
    [
        (TemporalEntity, (), {}, 'TemporalEntity("1970")'),
        (TemporalEntity, (2014,), {}, 'TemporalEntity("2014")'),
        (TemporalEntity, (2009, 12), {}, 'TemporalEntity("2009-12")'),
        (TemporalEntity, (1994, 12, 30), {}, 'TemporalEntity("1994-12-30")'),
        (
            TemporalEntity,
            (CET_tz.localize(datetime(1999, 1, 20, 22, 58, 17)),),  # noqa: DTZ001
            {},
            'TemporalEntity("1999-01-20T21:58:17Z")',
        ),
        (
            TemporalEntity,
            (datetime(1999, 1, 20, 22, 58, 17, tzinfo=UTC),),
            {},
            'TemporalEntity("1999-01-20T22:58:17Z")',
        ),
        (
            TemporalEntity,
            ("1994-12-30",),
            {},
            'TemporalEntity("1994-12-30")',
        ),
        (
            TemporalEntity,
            ("1999-01-20T22:58:17Z",),
            {},
            'TemporalEntity("1999-01-20T22:58:17Z")',
        ),
        (
            TemporalEntity,
            ("1999-01-20T22",),
            {},
            'TemporalEntity("1999-01-20T21:42:00Z")',
        ),
        (
            TemporalEntity,
            ("2016-06-10T21:42:24.76073899Z",),
            {},
            'TemporalEntity("2016-06-10T21:42:24Z")',
        ),
        (
            TemporalEntity,
            (date(2020, 3, 22),),
            {},
            'TemporalEntity("2020-03-22")',
        ),
        (
            TemporalEntity,
            (CET_tz.localize(datetime(2020, 3, 22, 14, 30, 58)),),  # noqa: DTZ001
            {},
            'TemporalEntity("2020-03-22T13:30:58Z")',
        ),
        (
            TemporalEntity,
            (PST_tz.localize(datetime(2020, 3, 22, 14, 30, 58)),),  # noqa: DTZ001
            {},
            'TemporalEntity("2020-03-22T21:30:58Z")',
        ),
        (
            TemporalEntity,
            (TemporalEntity(2004, 11),),
            {},
            'TemporalEntity("2004-11")',
        ),
        (
            YearMonthDayTime,
            (YearMonthDayTime(2004, 11, 21, 19, 59, tzinfo=UTC),),
            {},
            'YearMonthDayTime("2004-11-21T19:59:00Z")',
        ),
        (
            TemporalEntity,
            (CET_tz.localize(datetime(2004, 11, 19, 0, 0)),),  # noqa: DTZ001
            {"precision": TemporalEntityPrecision.DAY},
            'TemporalEntity("2004-11-19")',
        ),
        (
            Year,
            (CET_tz.localize(datetime(2004, 11, 19, 0, 0)),),  # noqa: DTZ001
            {"precision": TemporalEntityPrecision.YEAR},
            'Year("2004")',
        ),
    ],
    ids=[
        "empty",
        "year",
        "month",
        "day",
        "cet time",
        "utc time",
        "date string",
        "time string",
        "padded time",
        "nano seconds",
        "date",
        "datetime",
        "pacific time",
        "temporal entity",
        "sub class",
        "temporal entity with precision",
        "sub class with precision",
    ],
)
def test_temporal_entity_parsing(
    cls: type[TemporalEntity], args: tuple[Any], kwargs: dict[str, Any], expected: str
) -> None:
    temporal_entity = cls(*args, **kwargs)
    assert repr(temporal_entity) == expected


def test_temporal_entity_eq() -> None:
    assert TemporalEntity(2004) == TemporalEntity("2004")
    assert TemporalEntity(2004, 11) == TemporalEntity(2004, 11)
    assert TemporalEntity(2004, 11, 2) == "2004-11-02"
    assert TemporalEntity(2020, 3, 22, 14, 30, 58, 0, tzinfo=UTC) == datetime(
        2020, 3, 22, 14, 30, 58, 0, tzinfo=UTC
    )
    assert TemporalEntity(2005) != object()


def test_temporal_entity_gt() -> None:
    assert TemporalEntity(2004) > TemporalEntity("2003")
    assert TemporalEntity(2004, 11) < "2013-10-02"
    assert TemporalEntity(2004, 11) <= TemporalEntity(2004, 12)
    assert TemporalEntity(2020, 3, 22, 14, 30, 58, tzinfo=UTC) >= datetime(
        2020, 3, 22, 14, 29, tzinfo=UTC
    )

    with pytest.raises(NotImplementedError):
        assert TemporalEntity(2005) > object()


def test_temporal_entity_str() -> None:
    assert str(TemporalEntity(2004, 11, 26)) == "2004-11-26"
    assert str(YearMonth(2004, 11)) == "2004-11"
    assert str(YearMonthDay(2004, 11, 26)) == "2004-11-26"
    assert str(YearMonthDayTime(2018, 3, 2, 13, 0, 1)) == "2018-03-02T12:42:01Z"


def test_temporal_entity_repr() -> None:
    assert (
        repr(TemporalEntity(2018, 3, 2, 13, 0, 1, tzinfo=UTC))
        == 'TemporalEntity("2018-03-02T13:00:01Z")'
    )
    assert repr(Year("2022")) == 'Year("2022")'
    assert repr(YearMonth("2022-10")) == 'YearMonth("2022-10")'
    assert repr(YearMonthDay("2022-10-03")) == 'YearMonthDay("2022-10-03")'
    assert (
        repr(YearMonthDayTime("2018-03-02T12:00:01Z"))
        == 'YearMonthDayTime("2018-03-02T12:00:01Z")'
    )


class DummyModel(BaseModel):
    birthday: YearMonthDay


def test_temporal_entity_schema() -> None:
    assert DummyModel.model_json_schema() == {
        "properties": {
            "birthday": {
                "examples": ["2014-08-24"],
                "pattern": YEAR_MONTH_DAY_REGEX,
                "title": "YearMonthDay",
                "type": "string",
            }
        },
        "required": ["birthday"],
        "title": "DummyModel",
        "type": "object",
    }


DummyModel.model_json_schema()


def test_temporal_entity_serialization() -> None:
    person = DummyModel.model_validate({"birthday": "24th July 1999"})

    assert person.model_dump_json() == '{"birthday":"1999-07-24"}'



from typing import TYPE_CHECKING

from mex.common.models.base.model import BaseModel


class AdditiveRule(BaseModel, extra="forbid"):
    """Base rule to add values to merged items."""


class SubtractiveRule(BaseModel, extra="forbid"):
    """Base rule to subtract values from merged items."""


class PreventiveRule(BaseModel, extra="forbid"):
    """Base rule to prevent primary sources for fields of merged items."""


class RuleSet(BaseModel, extra="forbid"):
    """Base class for a set of an additive, subtractive and preventive rule."""

    if TYPE_CHECKING:  # pragma: no cover
        additive: AdditiveRule
        subtractive: SubtractiveRule
        preventive: PreventiveRule



from collections.abc import Callable
from typing import Any
from unittest.mock import MagicMock, Mock

import pytest
from ldap3 import Connection
from ldap3.core.exceptions import LDAPSocketOpenError
from pytest import MonkeyPatch

from mex.common.ldap.connector import LDAPConnector

PagedSearchResults = list[list[dict[str, Any]]]
LDAPMocker = Callable[[PagedSearchResults], None]


SAMPLE_PERSON_ATTRS = {
    "company": ["RKI"],
    "department": ["XY"],
    "departmentNumber": ["XY2"],
    "displayName": ["Sample, Sam"],
    "employeeID": ["1024"],
    "givenName": ["Sam"],
    "mail": ["SampleS@mail.tld"],
    "objectGUID": ["{00000000-0000-4000-8000-000000000000}"],
    "ou": ["XY"],
    "sAMAccountName": ["SampleS"],
    "sn": ["Sample"],
}

XY_FUNC_ACCOUNT_ATTRS = {
    "mail": ["XY@mail.tld"],
    "objectGUID": ["{00000000-0000-4000-8000-000000000044}"],
    "sAMAccountName": ["XY"],
    "ou": ["Funktion"],
}

XY2_FUNC_ACCOUNT_ATTRS = {
    "mail": ["XY2@mail.tld"],
    "objectGUID": ["{00000000-0000-4000-8000-000000000045}"],
    "sAMAccountName": ["XY2"],
    "ou": ["Funktion"],
}


@pytest.fixture
def ldap_mocker(monkeypatch: MonkeyPatch) -> LDAPMocker:
    """Patch the LDAP connector to return `SAMPLE_PERSON_ATTRS` from its connection."""

    def mocker(results: PagedSearchResults) -> None:
        def __init__(self: LDAPConnector) -> None:
            self._search_base = "DC=foo"
            self._connection = MagicMock(spec=Connection, extend=Mock())
            self._connection.extend.standard.paged_search = MagicMock(
                side_effect=[
                    [{"attributes": e} for e in entries] for entries in results
                ]
            )

        monkeypatch.setattr(LDAPConnector, "__init__", __init__)

    return mocker


@pytest.fixture(autouse=True)
def skip_ldap_integration_tests_on_connection_error(
    is_integration_test: bool,  # noqa: FBT001
) -> None:
    if is_integration_test:
        try:
            LDAPConnector.get()
        except LDAPSocketOpenError:
            pytest.skip("LDAP unavailable")



import json
from collections.abc import Iterable
from functools import lru_cache
from pathlib import Path

from mex.common.logging import logger
from mex.common.models import ExtractedOrganizationalUnit
from mex.common.organigram.models import OrganigramUnit
from mex.common.settings import BaseSettings
from mex.common.types import MergedOrganizationalUnitIdentifier


@lru_cache(maxsize=1)
def extract_organigram_units() -> list[OrganigramUnit]:
    """Extract organizational units from the organigram JSON file.

    Settings:
        organigram_path: Resolved path to the organigram file

    Returns:
        List of organigram units
    """
    settings = BaseSettings.get()
    with Path(settings.organigram_path).open() as fh:
        raw_units = json.load(fh)
    logger.info("extracted %s organigram units", len(raw_units))
    return [OrganigramUnit.model_validate(raw) for raw in raw_units]


def get_unit_synonyms(extracted_unit: ExtractedOrganizationalUnit) -> list[str]:
    """Generate synonyms for a unit using its name fields.

    Args:
        extracted_unit: Extracted organizational unit

    Returns:
        Sorted list of unique unit synonyms
    """
    return sorted(
        {
            extracted_unit.identifierInPrimarySource,
            *(
                text.value
                for texts in [
                    extracted_unit.name,
                    extracted_unit.shortName,
                    extracted_unit.alternativeName,
                ]
                for text in texts
            ),
        }
    )


def get_extracted_unit_by_synonyms(
    extracted_units: Iterable[ExtractedOrganizationalUnit],
) -> dict[str, list[ExtractedOrganizationalUnit]]:
    """Return a mapping from unit alt_label and label to their organizational units.

    Multiple units can share the same synonym, all will be included in the list.

    Args:
        extracted_units: Iterable of extracted units

    Returns:
        Mapping from unit synonyms to list of extracted units
    """
    synonym_dict: dict[str, list[ExtractedOrganizationalUnit]] = {}
    for extracted_unit in extracted_units:
        for synonym in get_unit_synonyms(extracted_unit):
            if synonym not in synonym_dict:
                synonym_dict[synonym] = []
            if extracted_unit not in synonym_dict[synonym]:
                synonym_dict[synonym].append(extracted_unit)
    return synonym_dict


def get_unit_merged_ids_by_synonyms(
    extracted_units: Iterable[ExtractedOrganizationalUnit],
) -> dict[str, list[MergedOrganizationalUnitIdentifier]]:
    """Return a mapping from unit alt_label and label to their merged IDs.

    Multiple units can share the same synonym, all will be included in the list.

    Args:
        extracted_units: Iterable of extracted units

    Returns:
        Mapping from unit synonyms to list of stableTargetIds
    """
    unit_dict = get_extracted_unit_by_synonyms(extracted_units)
    return {
        key: [unit.stableTargetId for unit in units] for key, units in unit_dict.items()
    }


def get_unit_merged_ids_by_emails(
    extracted_units: Iterable[ExtractedOrganizationalUnit],
) -> dict[str, list[MergedOrganizationalUnitIdentifier]]:
    """Return a mapping from unit emails to their merged IDs.

    Multiple units can share the same email, all will be included in the list.

    Args:
        extracted_units: Iterable of extracted units

    Returns:
        Mapping from lowercased `email` to list of stableTargetIds
    """
    email_dict: dict[str, list[MergedOrganizationalUnitIdentifier]] = {}
    for extracted_unit in extracted_units:
        for email in extracted_unit.email:
            lower_email = email.lower()
            if lower_email not in email_dict:
                email_dict[lower_email] = []
            if extracted_unit.stableTargetId not in email_dict[lower_email]:
                email_dict[lower_email].append(extracted_unit.stableTargetId)
    return email_dict



import re
from typing import Any
from unittest.mock import MagicMock
from uuid import UUID

import pytest
from ldap3.core.exceptions import LDAPSocketSendError
from pytest import MonkeyPatch

from mex.common.exceptions import MExError
from mex.common.ldap.connector import LDAPConnector
from mex.common.ldap.models import LDAPFunctionalAccount, LDAPPerson
from tests.ldap.conftest import (
    SAMPLE_PERSON_ATTRS,
    XY2_FUNC_ACCOUNT_ATTRS,
    XY_FUNC_ACCOUNT_ATTRS,
    LDAPMocker,
    PagedSearchResults,
)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("", ""),
        ("Tést@exämple.com", "Tést@exämple.com"),
        ("test*wildcard and spaces", "test*wildcard and spaces"),
        ("bad$chars#(here)!", "bad*chars*here*"),
    ],
)
def test_sanitize_value(value: str, expected: str) -> None:
    assert LDAPConnector._sanitize(value) == expected


def test_get_persons_mocked(ldap_mocker: LDAPMocker) -> None:
    ldap_mocker([[SAMPLE_PERSON_ATTRS]])
    connector = LDAPConnector.get()
    persons = connector.get_persons(surname="Sample", given_name="Kim")

    assert len(persons) == 1
    assert persons[0].model_dump(exclude_none=True) == {
        "company": "RKI",
        "department": "XY",
        "departmentNumber": "XY2",
        "displayName": "Sample, Sam",
        "employeeID": "1024",
        "givenName": ["Sam"],
        "mail": ["SampleS@mail.tld"],
        "objectGUID": UUID(int=0, version=4),
        "ou": ["XY"],
        "sAMAccountName": "SampleS",
        "sn": "Sample",
    }


@pytest.mark.parametrize(
    ("kwargs", "pattern"),
    [
        ({"surname": "müller"}, r".{37,}"),  # needs more than one guid to pass
        ({"surname": "nobody-has-this-name"}, r""),  # only empty list passes
    ],
    ids=[
        "common_surname",
        "nonexistent_person",
    ],
)
@pytest.mark.integration
def test_get_persons_ldap(kwargs: dict[str, Any], pattern: str) -> None:
    connector = LDAPConnector.get()
    persons = connector.get_persons(**kwargs)

    flat_result = ",".join(str(p.objectGUID) for p in persons)
    assert re.match(pattern, flat_result)


@pytest.mark.integration
def test_get_persons_or_functional_accounts_ldap() -> None:
    connector = LDAPConnector.get()
    assert len(connector.get_persons_or_functional_accounts(query="mex@rki.de")) == 1
    assert len(connector.get_persons_or_functional_accounts(query="*a*", limit=7)) == 7
    assert not connector.get_persons_or_functional_accounts(query="non-existent-bla")


def test_get_persons_or_functional_accounts_mocked(ldap_mocker: LDAPMocker) -> None:
    ldap_mocker([[XY_FUNC_ACCOUNT_ATTRS, SAMPLE_PERSON_ATTRS]])
    connector = LDAPConnector.get()
    functional_accounts = connector.get_persons_or_functional_accounts(query="XY")
    assert functional_accounts == [
        LDAPFunctionalAccount(
            displayName=None,
            mail=["XY@mail.tld"],
            objectGUID=UUID("00000000-0000-4000-8000-000000000044"),
            sAMAccountName="XY",
            ou=["Funktion"],
        ),
        LDAPPerson(
            displayName="Sample, Sam",
            mail=["SampleS@mail.tld"],
            objectGUID=UUID("00000000-0000-4000-8000-000000000000"),
            sAMAccountName="SampleS",
            company="RKI",
            department="XY",
            departmentNumber="XY2",
            employeeID="1024",
            givenName=["Sam"],
            ou=["XY"],
            sn="Sample",
        ),
    ]


@pytest.mark.parametrize(
    ("mail", "pattern"),
    [
        ("mex@rki.de", r".{36}"),  # exactly one guid
        ("non-existent@function.xyz", r"^$"),  # empty result
    ],
    ids=[
        "mex_functional_account",
        "nonexistent_functional_account",
    ],
)
@pytest.mark.integration
def test_get_functional_accounts_ldap(mail: str, pattern: str) -> None:
    connector = LDAPConnector.get()
    functional_accounts = connector.get_functional_accounts(mail=mail)

    flat_result = ",".join(str(p.objectGUID) for p in functional_accounts)
    assert re.match(pattern, flat_result)


def test_get_functional_accounts_mocked(ldap_mocker: LDAPMocker) -> None:
    ldap_mocker([[XY_FUNC_ACCOUNT_ATTRS]])
    connector = LDAPConnector.get()
    functional_accounts = connector.get_functional_accounts(mail="XY@mail.tld")

    assert len(functional_accounts) == 1
    assert functional_accounts[0].model_dump(exclude_none=True) == {
        "mail": ["XY@mail.tld"],
        "objectGUID": UUID("00000000-0000-4000-8000-000000000044"),
        "sAMAccountName": "XY",
        "ou": ["Funktion"],
    }


@pytest.mark.parametrize(
    ("search_results", "error_text"),
    [
        ([[]], "Cannot find AD person"),
        ([[SAMPLE_PERSON_ATTRS, SAMPLE_PERSON_ATTRS]], "Found multiple AD persons"),
    ],
)
def test_get_person_mocked_error(
    ldap_mocker: LDAPMocker, search_results: PagedSearchResults, error_text: str
) -> None:
    ldap_mocker(search_results)
    connector = LDAPConnector.get()
    with pytest.raises(MExError, match=error_text):
        connector.get_person(object_guid="whatever")


def test_get_person_mocked(ldap_mocker: LDAPMocker) -> None:
    ldap_mocker([[SAMPLE_PERSON_ATTRS]])
    connector = LDAPConnector.get()

    person = connector.get_person(object_guid=SAMPLE_PERSON_ATTRS["objectGUID"][0])

    expected = {
        "company": "RKI",
        "department": "XY",
        "departmentNumber": "XY2",
        "displayName": "Sample, Sam",
        "employeeID": "1024",
        "givenName": ["Sam"],
        "mail": ["SampleS@mail.tld"],
        "objectGUID": UUID("00000000-0000-4000-8000-000000000000"),
        "ou": ["XY"],
        "sAMAccountName": "SampleS",
        "sn": "Sample",
    }
    assert person.model_dump(exclude_none=True) == expected


@pytest.mark.parametrize(
    ("search_results", "error_text"),
    [
        ([[]], "Cannot find AD functional account"),
        (
            [[XY_FUNC_ACCOUNT_ATTRS, XY2_FUNC_ACCOUNT_ATTRS]],
            "Found multiple AD functional accounts",
        ),
    ],
)
def test_get_functional_account_mocked_error(
    ldap_mocker: LDAPMocker, search_results: PagedSearchResults, error_text: str
) -> None:
    ldap_mocker(search_results)
    connector = LDAPConnector.get()
    with pytest.raises(MExError, match=error_text):
        connector.get_functional_account(mail="whatever")


def test_functional_account_mocked(ldap_mocker: LDAPMocker) -> None:
    ldap_mocker([[XY_FUNC_ACCOUNT_ATTRS]])
    connector = LDAPConnector.get()

    unit = connector.get_functional_account(mail=XY_FUNC_ACCOUNT_ATTRS["mail"][0])

    expected = {
        "mail": ["XY@mail.tld"],
        "objectGUID": UUID("00000000-0000-4000-8000-000000000044"),
        "sAMAccountName": "XY",
        "ou": ["Funktion"],
    }
    assert unit.model_dump(exclude_none=True) == expected


def test_fetch_backoff_reconnect(monkeypatch: MonkeyPatch) -> None:
    # first connection raises ldap error
    first_connection = MagicMock(name="conn1")
    first_connection.extend.standard.paged_search = MagicMock(
        side_effect=LDAPSocketSendError("Simulated error")
    )
    # second connection returns valid content
    second_connection: MagicMock = MagicMock(name="conn2")
    second_connection.extend.standard.paged_search = MagicMock(
        return_value=[
            {
                "attributes": {
                    "sAMAccountName": "foo",
                    "objectGUID": "00000000-0000-4000-8000-000000000000",
                    "ou": ["Funktion"],
                }
            }
        ]
    )

    monkeypatch.setattr(
        LDAPConnector,
        "_setup_connection",
        MagicMock(side_effect=[first_connection, second_connection]),
    )
    connector = LDAPConnector.get()
    assert connector._connection is first_connection
    result = connector._fetch("(objectCategory=Person)", 1)
    assert result[0] == {
        "sAMAccountName": "foo",
        "objectGUID": "00000000-0000-4000-8000-000000000000",
        "ou": ["Funktion"],
    }
    assert connector._connection is second_connection



from mex.common.testing import Joker


def test_joker_eq() -> None:
    assert Joker() == None  # noqa: E711
    assert Joker() == 1
    assert {"foo": Joker()} == {"foo": ["bar", Joker()]}


def test_joker_repr() -> None:
    joker = Joker()
    assert {"question": joker} == {"question": "Why so serious?"}
    assert repr(joker) == "'Why so serious?'"



from unittest.mock import MagicMock, Mock

import pytest
import requests
from pytest import MonkeyPatch

from mex.common.backend_api.connector import BackendApiConnector
from mex.common.identity import Identity
from mex.common.identity.backend_api import BackendApiIdentityProvider
from mex.common.models import ExtractedContactPoint
from mex.common.types import MergedPrimarySourceIdentifier


@pytest.fixture
def mocked_backend_connector(monkeypatch: MonkeyPatch) -> MagicMock:
    mocked_session = MagicMock(spec=requests.Session)
    mocked_session.request = MagicMock(
        return_value=Mock(spec=requests.Response, status_code=200)
    )
    mocked_session.headers = {}

    def set_mocked_session(self: BackendApiConnector) -> None:
        self.session = mocked_session

    monkeypatch.setattr(BackendApiConnector, "_set_session", set_mocked_session)
    return mocked_session


def test_assign_mocked(
    mocked_backend_connector: requests.Session,
) -> None:
    mocked_data = {
        "identifier": MergedPrimarySourceIdentifier.generate(seed=962),
        "hadPrimarySource": MergedPrimarySourceIdentifier.generate(seed=961),
        "identifierInPrimarySource": "test",
        "stableTargetId": MergedPrimarySourceIdentifier.generate(seed=963),
    }
    mocked_response = Mock(spec=requests.Response)
    mocked_response.status_code = 200
    mocked_response.json = MagicMock(return_value=mocked_data)
    mocked_backend_connector.request = MagicMock(return_value=mocked_response)  # type: ignore[method-assign]

    provider = BackendApiIdentityProvider.get()
    identity_first = provider.assign(
        had_primary_source=MergedPrimarySourceIdentifier.generate(seed=961),
        identifier_in_primary_source="test",
    )

    identity = Identity.model_validate(identity_first)

    identity_first_assignment = identity.model_dump()

    assert identity_first_assignment == mocked_data

    identity_second = provider.assign(
        had_primary_source=MergedPrimarySourceIdentifier.generate(seed=961),
        identifier_in_primary_source="test",
    )
    identity_second_assignment = identity_second.model_dump()

    assert identity_second_assignment == identity_first_assignment


def test_fetch_mocked(
    mocked_backend_connector: requests.Session,
) -> None:
    mocked_data = {
        "items": [
            {
                "identifier": MergedPrimarySourceIdentifier.generate(seed=962),
                "hadPrimarySource": MergedPrimarySourceIdentifier.generate(seed=961),
                "identifierInPrimarySource": "test",
                "stableTargetId": MergedPrimarySourceIdentifier.generate(seed=963),
            }
        ],
        "total": 1,
    }

    mocked_response = Mock(spec=requests.Response)
    mocked_response.status_code = 200
    mocked_response.json = MagicMock(return_value=mocked_data)
    mocked_backend_connector.request = MagicMock(return_value=mocked_response)  # type: ignore[method-assign]

    provider = BackendApiIdentityProvider.get()

    contact_point = ExtractedContactPoint(
        hadPrimarySource=MergedPrimarySourceIdentifier.generate(seed=961),
        identifierInPrimarySource="test",
        email=["test@test.de"],
    )

    identities = provider.fetch(stable_target_id=contact_point.stableTargetId)
    assert identities == [
        Identity(
            stableTargetId=mocked_data["items"][0]["stableTargetId"],  # type: ignore[index]
            identifier=mocked_data["items"][0]["identifier"],  # type: ignore[index]
            hadPrimarySource=contact_point.hadPrimarySource,
            identifierInPrimarySource=contact_point.identifierInPrimarySource,
        )
    ]

    identities = provider.fetch(
        had_primary_source=contact_point.hadPrimarySource,
        identifier_in_primary_source=contact_point.identifierInPrimarySource,
    )
    assert identities == [
        Identity(
            stableTargetId=mocked_data["items"][0]["stableTargetId"],  # type: ignore[index]
            identifier=mocked_data["items"][0]["identifier"],  # type: ignore[index]
            hadPrimarySource=contact_point.hadPrimarySource,
            identifierInPrimarySource=contact_point.identifierInPrimarySource,
        )
    ]


def test_fetch_mocked_empty(
    mocked_backend_connector: requests.Session,
) -> None:
    mocked_response = Mock(spec=requests.Response)
    mocked_response.status_code = 200
    mocked_response.json = MagicMock(return_value={"items": [], "total": 0})
    mocked_backend_connector.request = MagicMock(return_value=mocked_response)  # type: ignore[method-assign]

    provider = BackendApiIdentityProvider.get()

    provider.fetch()

    mocked_backend_connector.request.assert_called_with(
        "GET",
        "http://localhost:8080/v0/identity",
        {
            "hadPrimarySource": None,
            "identifierInPrimarySource": None,
            "stableTargetId": None,
        },
        timeout=10,
        headers={"Accept": "application/json", "User-Agent": "rki/mex"},
    )



from mex.common.organigram.helpers import build_child_map, find_descendants
from mex.common.organigram.models import OrganigramUnit


def test_build_child_map(
    child_unit: OrganigramUnit, parent_unit: OrganigramUnit
) -> None:
    test_map = build_child_map(
        [child_unit, child_unit, parent_unit, child_unit, parent_unit]
    )

    assert test_map == {"parent-unit": ["child-unit", "child-unit", "child-unit"]}


def test_find_descendants(
    child_unit: OrganigramUnit, parent_unit: OrganigramUnit
) -> None:
    child_ids = find_descendants(
        [child_unit, child_unit, child_unit, parent_unit, parent_unit],
        str(parent_unit.identifier),
    )

    assert child_ids == [child_unit.identifier]
